import duckdb
import pytest
import tempfile
import os
import gc
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestPandasUnregister(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_pandas_unregister1(self, duckdb_cursor, pandas):
        df = pandas.DataFrame([[1, 2, 3], [4, 5, 6]])
        connection = duckdb.connect(":memory:")
        connection.register("dataframe", df)

        df2 = connection.execute("SELECT * FROM dataframe;").fetchdf()
        connection.unregister("dataframe")
        with pytest.raises(duckdb.CatalogException, match='Table with name dataframe does not exist'):
            connection.execute("SELECT * FROM dataframe;").fetchdf()
        with pytest.raises(duckdb.CatalogException, match='View with name dataframe does not exist'):
            connection.execute("DROP VIEW dataframe;")
        connection.execute("DROP VIEW IF EXISTS dataframe;")

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_pandas_unregister2(self, duckdb_cursor, pandas):
        fd, db = tempfile.mkstemp()
        os.close(fd)
        os.remove(db)

        connection = duckdb.connect(db)
        df = pandas.DataFrame([[1, 2, 3], [4, 5, 6]])

        connection.register("dataframe", df)
        connection.unregister("dataframe")  # Attempting to unregister.
        connection.close()

        # Reconnecting while DataFrame still in mem.
        connection = duckdb.connect(db)
        assert len(connection.execute("PRAGMA show_tables;").fetchall()) == 0

        with pytest.raises(duckdb.CatalogException, match='Table with name dataframe does not exist'):
            connection.execute("SELECT * FROM dataframe;").fetchdf()

        connection.close()

        del df
        gc.collect()

        # Reconnecting after DataFrame freed.
        connection = duckdb.connect(db)
        assert len(connection.execute("PRAGMA show_tables;").fetchall()) == 0
        with pytest.raises(duckdb.CatalogException, match='Table with name dataframe does not exist'):
            connection.execute("SELECT * FROM dataframe;").fetchdf()
        connection.close()



import duckdb
import datetime
import numpy as np
import pytest
import copy
from conftest import NumpyPandas, ArrowPandas

NULL = None


def check_equal(conn, df, reference_query):
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute(reference_query)
    res = duckdb_conn.query('SELECT * FROM tbl').fetchall()
    df_res = duckdb_conn.query('SELECT * FROM tbl').df()
    out = conn.sql("SELECT * FROM df").fetchall()
    assert res == out


def create_reference_query(data):
    query = "CREATE TABLE tbl AS SELECT " + str(data).replace("None", "NULL")
    return query


class TestDFRecursiveNested(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_of_structs(self, duckdb_cursor, pandas):
        data = [[{'a': 5}, NULL, {'a': NULL}], NULL, [{'a': 5}, NULL, {'a': NULL}]]
        reference_query = create_reference_query(data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_of_map(self, duckdb_cursor, pandas):
        # LIST(MAP(VARCHAR, VARCHAR))
        data = [
            [{'key': [5], 'value': [NULL]}, NULL, {'key': [], 'value': []}],
            NULL,
            [NULL, {'key': [3, 2, 4], 'value': [NULL, 'a', NULL]}, {'key': ['a', 'b', 'c'], 'value': [1, 2, 3]}],
        ]
        reference_data = [
            [{'key': ['5'], 'value': [NULL]}, NULL, {'key': [], 'value': []}],
            NULL,
            [
                NULL,
                {'key': ['3', '2', '4'], 'value': [NULL, 'a', NULL]},
                {'key': ['a', 'b', 'c'], 'value': ['1', '2', '3']},
            ],
        ]
        reference_query = create_reference_query(reference_data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_recursive_list(self, duckdb_cursor, pandas):
        # LIST(LIST(LIST(LIST(INTEGER))))
        data = [[[[3, NULL, 5], NULL], NULL, [[5, -20, NULL]]], NULL, [[[NULL]], [[]], NULL]]
        reference_query = create_reference_query(data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_recursive_struct(self, duckdb_cursor, pandas):
        # STRUCT(STRUCT(STRUCT(LIST)))
        data = {
            'A': {'a': {'1': [1, 2, 3]}, 'b': NULL, 'c': {'1': NULL}},
            'B': {'a': {'1': [1, NULL, 3]}, 'b': NULL, 'c': {'1': NULL}},
        }
        reference_query = create_reference_query(data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_recursive_map(self, duckdb_cursor, pandas):
        # MAP(
        # 	MAP(
        # 		INTEGER,
        # 		MAP(INTEGER)
        # 	),
        # 	INTEGER
        # )
        data = {
            'key': [
                {'key': [5, 6, 7], 'value': [{'key': [8], 'value': [NULL]}, NULL, {'key': [9], 'value': ['a']}]},
                {'key': [], 'value': []},
            ],
            'value': [1, 2],
        }
        reference_query = create_reference_query(data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_recursive_stresstest(self, duckdb_cursor, pandas):
        # LIST(
        # 	STRUCT(
        # 		MAP(
        # 			STRUCT(
        # 				LIST(
        # 					INTEGER
        # 				)
        # 			)
        # 			LIST(
        # 				STRUCT(
        # 					VARCHAR
        # 				)
        # 			)
        # 		)
        # 	)
        # )
        data = [
            {
                'a': {
                    'key': [
                        {'1': [5, 4, 3], '2': [8, 7, 6], '3': [1, 2, 3]},
                        {'1': [], '2': NULL, '3': [NULL, 0, NULL]},
                    ],
                    'value': [[{'A': 'abc', 'B': 'def', 'C': NULL}], [NULL]],
                },
                'b': NULL,
                'c': {'key': [], 'value': []},
            }
        ]
        reference_query = create_reference_query(data)
        df = pandas.DataFrame([{'a': data}])
        check_equal(duckdb_cursor, df, reference_query)



import duckdb
import numpy


class TestBoolean(object):
    def test_bool(self, duckdb_cursor):
        duckdb_cursor.execute("SELECT TRUE")
        results = duckdb_cursor.fetchall()
        assert results[0][0] == True



import os

import duckdb
from pytest import raises
import pytest


def test_extension_loading(require):
    if not os.getenv('DUCKDB_PYTHON_TEST_EXTENSION_REQUIRED', False):
        return
    extensions_list = ['json', 'excel', 'httpfs', 'tpch', 'tpcds', 'icu', 'fts']
    for extension in extensions_list:
        connection = require(extension)
        assert connection is not None


def test_install_non_existent_extension():
    conn = duckdb.connect()
    conn.execute("set custom_extension_repository = 'http://example.com'")

    with raises(duckdb.IOException) as exc:
        conn.install_extension('non-existent')

    if not isinstance(exc, duckdb.HTTPException):
        pytest.skip(reason='This test does not throw an HTTPException, only an IOException')
    value = exc.value

    assert value.status_code == 404
    assert value.reason == 'Not Found'
    assert 'Example Domain' in value.body
    assert 'Content-Length' in value.headers



import duckdb
import os
import datetime
import pytest

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")
np = pytest.importorskip("numpy")


class TestArrowIntegration(object):
    def test_parquet_roundtrip(self, duckdb_cursor):
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        cols = 'id, first_name, last_name, email, gender, ip_address, cc, country, birthdate, salary, title, comments'

        # TODO timestamp

        userdata_parquet_table = pq.read_table(parquet_filename)
        userdata_parquet_table.validate(full=True)
        rel_from_arrow = duckdb.arrow(userdata_parquet_table).project(cols).arrow()
        rel_from_arrow.validate(full=True)

        rel_from_duckdb = duckdb_cursor.from_parquet(parquet_filename).project(cols).arrow()
        rel_from_duckdb.validate(full=True)

        # batched version, lets use various values for batch size
        for i in [7, 51, 99, 100, 101, 500, 1000, 2000]:
            userdata_parquet_table2 = pa.Table.from_batches(userdata_parquet_table.to_batches(i))
            assert userdata_parquet_table.equals(userdata_parquet_table2, check_metadata=True)

            rel_from_arrow2 = duckdb.arrow(userdata_parquet_table2).project(cols).arrow()
            rel_from_arrow2.validate(full=True)

            assert rel_from_arrow.equals(rel_from_arrow2, check_metadata=True)
            assert rel_from_arrow.equals(rel_from_duckdb, check_metadata=True)

    def test_unsigned_roundtrip(self, duckdb_cursor):
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'unsigned.parquet')
        cols = 'a, b, c, d'

        unsigned_parquet_table = pq.read_table(parquet_filename)
        unsigned_parquet_table.validate(full=True)
        rel_from_arrow = duckdb.arrow(unsigned_parquet_table).project(cols).arrow()
        rel_from_arrow.validate(full=True)

        rel_from_duckdb = duckdb_cursor.from_parquet(parquet_filename).project(cols).arrow()
        rel_from_duckdb.validate(full=True)

        assert rel_from_arrow.equals(rel_from_duckdb, check_metadata=True)

        duckdb_cursor.execute(
            "select NULL c_null, (c % 4 = 0)::bool c_bool, (c%128)::tinyint c_tinyint, c::smallint*1000::INT c_smallint, c::integer*100000 c_integer, c::bigint*1000000000000 c_bigint, c::float c_float, c::double c_double, 'c_' || c::string c_string from (select case when range % 2 == 0 then range else null end as c from range(-10000, 10000)) sq"
        )
        arrow_result = duckdb_cursor.fetch_arrow_table()
        arrow_result.validate(full=True)
        arrow_result.combine_chunks()
        arrow_result.validate(full=True)

        round_tripping = duckdb.from_arrow(arrow_result).to_arrow_table()
        round_tripping.validate(full=True)

        assert round_tripping.equals(arrow_result, check_metadata=True)

    def test_decimals_roundtrip(self, duckdb_cursor):
        duckdb_cursor.execute("CREATE TABLE test (a DECIMAL(4,2), b DECIMAL(9,2), c DECIMAL (18,2), d DECIMAL (30,2))")

        duckdb_cursor.execute("INSERT INTO  test VALUES (1.11,1.11,1.11,1.11),(NULL,NULL,NULL,NULL)")

        true_result = duckdb_cursor.execute("SELECT sum(a), sum(b), sum(c),sum(d) from test").fetchall()

        duck_tbl = duckdb_cursor.table("test")

        duck_from_arrow = duckdb_cursor.from_arrow(duck_tbl.arrow())

        duck_from_arrow.create("testarrow")

        arrow_result = duckdb_cursor.execute("SELECT sum(a), sum(b), sum(c),sum(d) from testarrow").fetchall()

        assert arrow_result == true_result

        arrow_result = duckdb_cursor.execute(
            "SELECT typeof(a), typeof(b), typeof(c),typeof(d) from testarrow"
        ).fetchone()

        assert arrow_result[0] == 'DECIMAL(4,2)'
        assert arrow_result[1] == 'DECIMAL(9,2)'
        assert arrow_result[2] == 'DECIMAL(18,2)'
        assert arrow_result[3] == 'DECIMAL(30,2)'

        # Lets also test big number comming from arrow land
        data = pa.array(np.array([9999999999999999999999999999999999]), type=pa.decimal128(38, 0))
        arrow_tbl = pa.Table.from_arrays([data], ['a'])
        duckdb_cursor.from_arrow(arrow_tbl).create("bigdecimal")
        result = duckdb_cursor.execute('select * from bigdecimal')
        assert result.fetchone()[0] == 9999999999999999999999999999999999

    def test_intervals_roundtrip(self, duckdb_cursor):
        # test for import from apache arrow
        expected_value = pa.MonthDayNano(
            [
                2,
                8,
                (
                    datetime.timedelta(seconds=1, microseconds=1, milliseconds=1, minutes=1, hours=1)
                    // datetime.timedelta(microseconds=1)
                )
                * 1000,
            ]
        )
        arr = [expected_value]

        data = pa.array(arr, pa.month_day_nano_interval())
        arrow_tbl = pa.Table.from_arrays([data], ['a'])
        duckdb_cursor.from_arrow(arrow_tbl).create("intervaltbl")
        duck_arrow_tbl = duckdb_cursor.table("intervaltbl").arrow()['a']

        assert duck_arrow_tbl[0].value == expected_value

        # test for select interval from duckdb
        duckdb_cursor.execute("CREATE TABLE test (a INTERVAL)")
        duckdb_cursor.execute("INSERT INTO  test VALUES (INTERVAL 1 YEAR + INTERVAL 1 DAY + INTERVAL 1 SECOND)")
        expected_value = pa.MonthDayNano([12, 1, 1000000000])
        duck_tbl_arrow = duckdb_cursor.table("test").arrow()['a']
        assert duck_tbl_arrow[0].value.months == expected_value.months
        assert duck_tbl_arrow[0].value.days == expected_value.days
        assert duck_tbl_arrow[0].value.nanoseconds == expected_value.nanoseconds

    def test_null_intervals_roundtrip(self, duckdb_cursor):
        # test for null interval
        expected_value = pa.MonthDayNano(
            [
                2,
                8,
                (
                    datetime.timedelta(seconds=1, microseconds=1, milliseconds=1, minutes=1, hours=1)
                    // datetime.timedelta(microseconds=1)
                )
                * 1000,
            ]
        )
        arr = [None, expected_value]
        data = pa.array(arr, pa.month_day_nano_interval())
        arrow_tbl = pa.Table.from_arrays([data], ['a'])
        duckdb_cursor.from_arrow(arrow_tbl).create("intervalnulltbl")
        duckdb_tbl_arrow = duckdb_cursor.table("intervalnulltbl").arrow()['a']

        assert duckdb_tbl_arrow[0].value == None
        assert duckdb_tbl_arrow[1].value == expected_value

    def test_nested_interval_roundtrip(self, duckdb_cursor):
        # Dictionary
        indices = pa.array([0, 1, 0, 1, 2, 1, 0, 2])
        first_value = pa.MonthDayNano([0, 1, 2000000])
        second_value = pa.MonthDayNano([90, 12, 0])
        dictionary = pa.array([first_value, second_value, None])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        duckdb_cursor.from_arrow(arrow_table).create("dictionarytbl")
        duckdb_tbl_arrow = duckdb_cursor.table("dictionarytbl").arrow()['a']

        assert duckdb_tbl_arrow[0].value == first_value
        assert duckdb_tbl_arrow[1].value == second_value
        assert duckdb_tbl_arrow[2].value == first_value
        assert duckdb_tbl_arrow[3].value == second_value
        assert duckdb_tbl_arrow[4].value == None
        assert duckdb_tbl_arrow[5].value == second_value
        assert duckdb_tbl_arrow[6].value == first_value
        assert duckdb_tbl_arrow[7].value == None

        # List
        query = duckdb_cursor.sql(
            "SELECT a from (select list_value(INTERVAL 3 MONTHS, INTERVAL 5 DAYS, INTERVAL 10 SECONDS, NULL) as a) as t"
        ).arrow()['a']
        assert query[0][0].value == pa.MonthDayNano([3, 0, 0])
        assert query[0][1].value == pa.MonthDayNano([0, 5, 0])
        assert query[0][2].value == pa.MonthDayNano([0, 0, 10000000000])
        assert query[0][3].value == None

        # Struct
        query = "SELECT a from (SELECT STRUCT_PACK(a := INTERVAL 1 MONTHS, b := INTERVAL 10 DAYS, c:= INTERVAL 20 SECONDS) as a) as t"
        true_answer = duckdb_cursor.sql(query).fetchall()
        from_arrow = duckdb_cursor.from_arrow(duckdb_cursor.sql(query).arrow()).fetchall()
        assert true_answer[0][0]['a'] == from_arrow[0][0]['a']
        assert true_answer[0][0]['b'] == from_arrow[0][0]['b']
        assert true_answer[0][0]['c'] == from_arrow[0][0]['c']

    def test_min_max_interval_roundtrip(self, duckdb_cursor):
        interval_min_value = pa.MonthDayNano([0, 0, 0])
        interval_max_value = pa.MonthDayNano([2147483647, 2147483647, 9223372036854775000])
        data = pa.array([interval_min_value, interval_max_value], pa.month_day_nano_interval())
        arrow_tbl = pa.Table.from_arrays([data], ['a'])
        duckdb_cursor.from_arrow(arrow_tbl).create("intervalminmaxtbl")

        duck_arrow_tbl = duckdb_cursor.table("intervalminmaxtbl").arrow()['a']
        assert duck_arrow_tbl[0].value == pa.MonthDayNano([0, 0, 0])
        assert duck_arrow_tbl[1].value == pa.MonthDayNano([2147483647, 2147483647, 9223372036854775000])

    def test_duplicate_column_names(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        df_a = pd.DataFrame({'join_key': [1, 2, 3], 'col_a': ['a', 'b', 'c']})
        df_b = pd.DataFrame({'join_key': [1, 3, 4], 'col_a': ['x', 'y', 'z']})

        res = duckdb_cursor.execute(
            """
            SELECT *
                FROM
                    df_a table1
                JOIN
                    df_b table2 ON
                table1.join_key = table2.join_key
        """
        ).arrow()
        assert res.schema.names == ['join_key', 'col_a', 'join_key', 'col_a']

    def test_strings_roundtrip(self, duckdb_cursor):
        duckdb_cursor.execute("CREATE TABLE test (a varchar)")

        # Test Small, Null and Very Big String
        for i in range(0, 1000):
            duckdb_cursor.execute(
                "INSERT INTO  test VALUES ('Matt Damon'),(NULL), ('Jeffffreeeey Jeeeeef Baaaaaaazos'), ('X-Content-Type-Options')"
            )

        true_result = duckdb_cursor.execute("SELECT * from test").fetchall()

        duck_tbl = duckdb_cursor.table("test")

        duck_from_arrow = duckdb_cursor.from_arrow(duck_tbl.arrow())

        duck_from_arrow.create("testarrow")

        arrow_result = duckdb_cursor.execute("SELECT * from testarrow").fetchall()

        assert arrow_result == true_result



import duckdb
import pytest


tf = pytest.importorskip('tensorflow')


def test_tf():
    con = duckdb.connect()

    con.execute("create table t( a integer, b integer)")
    con.execute("insert into t values (1,2), (3,4)")

    # Test from connection
    duck_tf = con.execute("select * from t").tf()
    duck_numpy = con.sql("select * from t").fetchnumpy()
    tf.math.equal(duck_tf['a'], tf.convert_to_tensor(duck_numpy['a']))
    tf.math.equal(duck_tf['b'], tf.convert_to_tensor(duck_numpy['b']))

    # Test from relation
    duck_tf = con.sql("select * from t").tf()
    tf.math.equal(duck_tf['a'], tf.convert_to_tensor(duck_numpy['a']))
    tf.math.equal(duck_tf['b'], tf.convert_to_tensor(duck_numpy['b']))

    # Test all Numeric Types
    numeric_types = ['TINYINT', 'SMALLINT', 'BIGINT', 'HUGEINT', 'FLOAT', 'DOUBLE', 'DECIMAL(4,1)', 'UTINYINT']

    for supported_type in numeric_types:
        con = duckdb.connect()
        con.execute(f"create table t( a {supported_type} , b {supported_type})")
        con.execute("insert into t values (1,2), (3,4)")
        duck_tf = con.sql("select * from t").tf()
        duck_numpy = con.sql("select * from t").fetchnumpy()
        tf.math.equal(duck_tf['a'], tf.convert_to_tensor(duck_numpy['a']))
        tf.math.equal(duck_tf['b'], tf.convert_to_tensor(duck_numpy['b']))



import duckdb
import numpy


def check_result(duckdb_cursor, value, type):
    duckdb_cursor.execute("SELECT " + str(value) + "::" + type)
    results = duckdb_cursor.fetchall()
    assert results[0][0] == value


class TestNumeric(object):
    def test_numeric_results(self, duckdb_cursor):
        check_result(duckdb_cursor, 1, "TINYINT")
        check_result(duckdb_cursor, 1, "SMALLINT")
        check_result(duckdb_cursor, 1, "FLOAT")



import duckdb
import os
import datetime
import pytest
import pandas as pd
import io


class TestPandasStringNull(object):
    def test_pandas_string_null(self, duckdb_cursor):
        csv = u'''what,is_control,is_test
,0,0
foo,1,0'''
        df = pd.read_csv(io.StringIO(csv))
        duckdb_cursor.register("c", df)
        duckdb_cursor.execute('select what, count(*) from c group by what')
        df_result = duckdb_cursor.fetchdf()
        assert True  # Should not crash ^^



import duckdb

try:
    import pyarrow as pa
    import pyarrow.dataset as ds

    can_run = True
except:
    can_run = False


class TestArrowTypes(object):
    def test_null_type(self, duckdb_cursor):
        if not can_run:
            return
        schema = pa.schema([("data", pa.null())])
        inputs = [pa.array([None, None, None], type=pa.null())]
        arrow_table = pa.Table.from_arrays(inputs, schema=schema)
        duckdb_conn = duckdb.connect()
        duckdb_conn.register("testarrow", arrow_table)
        rel = duckdb.from_arrow(arrow_table).arrow()
        # We turn it to an array of int32 nulls
        schema = pa.schema([("data", pa.int32())])
        inputs = [pa.array([None, None, None], type=pa.null())]
        arrow_table = pa.Table.from_arrays(inputs, schema=schema)

        assert rel['data'] == arrow_table['data']



import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestParameterList(object):
    def test_bool(self, duckdb_cursor):
        conn = duckdb.connect()
        conn.execute("create table bool_table (a bool)")
        conn.execute("insert into bool_table values (TRUE)")
        res = conn.execute("select count(*) from bool_table where a =?", [True])
        assert res.fetchone()[0] == 1

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_exception(self, duckdb_cursor, pandas):
        conn = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.execute("create table bool_table (a bool)")
        conn.execute("insert into bool_table values (TRUE)")
        with pytest.raises(duckdb.NotImplementedException, match='Unable to transform'):
            res = conn.execute("select count(*) from bool_table where a =?", [df_in])

    def test_explicit_nan_param(self):
        con = duckdb.default_connection
        res = con.execute('select isnan(cast(? as double))', (float("nan"),))
        assert res.fetchone()[0] == True



import duckdb
import pytest
from duckdb.typing import INTEGER, VARCHAR, TIMESTAMP
from duckdb import Expression, ConstantExpression, ColumnExpression, StarExpression, FunctionExpression, CaseExpression
from duckdb.value.constant import Value, IntegerValue
import datetime


@pytest.fixture(scope='function')
def filter_rel():
    con = duckdb.connect()
    rel = con.sql(
        """
        select * from (VALUES
            (1, 'a'),
            (2, 'b'),
            (1, 'b'),
            (3, 'c'),
            (4, 'a')
        ) tbl(a, b)
    """
    )
    yield rel


class TestExpression(object):
    def test_constant_expression(self):
        con = duckdb.connect()

        val = Value(5, INTEGER)

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c
        """
        )

        constant = ConstantExpression(val)

        rel = rel.select(constant)
        res = rel.fetchall()
        assert res == [(5,)]

    def test_column_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c
        """
        )
        column = ColumnExpression('a')
        rel2 = rel.select(column)
        res = rel2.fetchall()
        assert res == [(1,)]

        column = ColumnExpression('d')
        with pytest.raises(duckdb.BinderException, match='Referenced column "d" not found'):
            rel2 = rel.select(column)

    def test_column_expression_explain(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select 'unused'
        """
        )
        rel = rel.select(
            ConstantExpression("a").alias('c0'),
            ConstantExpression(42).alias('c1'),
            ConstantExpression(None).alias('c2'),
        )
        res = rel.explain()
        assert 'c0' in res
        assert 'c1' in res
        # 'c2' is not in the explain result because it shows NULL instead
        assert 'NULL' in res
        res = rel.fetchall()
        assert res == [('a', 42, None)]

    def test_column_expression_table(self):
        con = duckdb.connect()

        con.execute(
            """
            CREATE TABLE tbl as FROM (
                VALUES
                    ('a', 'b', 'c'),
                    ('d', 'e', 'f'),
                    ('g', 'h', 'i')
            ) t(c0, c1, c2)
        """
        )

        rel = con.table('tbl')
        rel2 = rel.select('c0', 'c1', 'c2')
        res = rel2.fetchall()
        assert res == [('a', 'b', 'c'), ('d', 'e', 'f'), ('g', 'h', 'i')]

    def test_column_expression_view(self):
        con = duckdb.connect()
        con.execute(
            """
            CREATE TABLE tbl as FROM (
                VALUES
                    ('a', 'b', 'c'),
                    ('d', 'e', 'f'),
                    ('g', 'h', 'i')
            ) t(c0, c1, c2)
        """
        )
        con.execute(
            """
            CREATE VIEW v1 as select c0 as c3, c2 as c4 from tbl;
        """
        )
        rel = con.view('v1')
        rel2 = rel.select('c3', 'c4')
        res = rel2.fetchall()
        assert res == [('a', 'c'), ('d', 'f'), ('g', 'i')]

    def test_column_expression_replacement_scan(self):
        con = duckdb.connect()

        pd = pytest.importorskip("pandas")
        df = pd.DataFrame({'a': [42, 43, 0], 'b': [True, False, True], 'c': [23.123, 623.213, 0.30234]})
        rel = con.sql("select * from df")
        rel2 = rel.select('a', 'b')
        res = rel2.fetchall()
        assert res == [(42, True), (43, False), (0, True)]

    def test_add_operator(self):
        con = duckdb.connect()

        val = Value(5, INTEGER)

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c
        """
        )

        constant = ConstantExpression(val)
        col = ColumnExpression('b')
        expr = col + constant

        rel = rel.select(expr, expr)
        res = rel.fetchall()
        assert res == [(7, 7)]

    def test_binary_function_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                5 as b
        """
        )
        function = FunctionExpression("-", ColumnExpression('b'), ColumnExpression('a'))
        rel2 = rel.select(function)
        res = rel2.fetchall()
        assert res == [(4,)]

    def test_negate_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select 5 as a
        """
        )
        col = ColumnExpression('a')
        col = -col
        rel = rel.select(col)
        res = rel.fetchall()
        assert res == [(-5,)]

    def test_subtract_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                3 as a,
                1 as b
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = col1 - col2
        rel = rel.select(expr)
        res = rel.fetchall()
        assert res == [(2,)]

    def test_multiply_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                3 as a,
                2 as b
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = col1 * col2
        rel = rel.select(expr)
        res = rel.fetchall()
        assert res == [(6,)]

    def test_division_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                5 as a,
                2 as b
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = col1 / col2
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(2.5,)]

        expr = col1 // col2
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(2,)]

    def test_modulus_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                5 as a,
                2 as b
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = col1 % col2
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(1,)]

    def test_power_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                5 as a,
                2 as b
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = col1**col2
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(25,)]

    def test_equality_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                5 as a,
                2 as b,
                5 as c
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        col3 = ColumnExpression('c')
        expr1 = col1 == col2
        expr2 = col1 == col3
        rel2 = rel.select(expr1, expr2)
        res = rel2.fetchall()
        assert res == [(False, True)]

    def test_inequality_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                5 as a,
                2 as b,
                5 as c
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        col3 = ColumnExpression('c')
        expr1 = col1 != col2
        expr2 = col1 != col3
        rel2 = rel.select(expr1, expr2)
        res = rel2.fetchall()
        assert res == [(True, False)]

    def test_comparison_expressions(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c,
                3 as d
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        col3 = ColumnExpression('c')
        col4 = ColumnExpression('d')

        # Greater than
        expr1 = col1 > col2
        expr2 = col2 > col1
        expr3 = col3 > col4
        rel2 = rel.select(expr1, expr2, expr3)
        res = rel2.fetchall()
        assert res == [(False, True, False)]

        # Greater than or equal
        expr1 = col1 >= col2
        expr2 = col2 >= col1
        expr3 = col3 >= col4
        rel2 = rel.select(expr1, expr2, expr3)
        res = rel2.fetchall()
        assert res == [(False, True, True)]

        # Less than
        expr1 = col1 < col2
        expr2 = col2 < col1
        expr3 = col3 < col4
        rel2 = rel.select(expr1, expr2, expr3)
        res = rel2.fetchall()
        assert res == [(True, False, False)]

        # Less than or equal
        expr1 = col1 <= col2
        expr2 = col2 <= col1
        expr3 = col3 <= col4
        rel2 = rel.select(expr1, expr2, expr3)
        res = rel2.fetchall()
        assert res == [(True, False, True)]

    def test_expression_alias(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select 1 as a
        """
        )
        col = ColumnExpression('a')
        col = col.alias('b')

        rel2 = rel.select(col)
        assert rel2.columns == ['b']

    def test_star_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                2 as b
        """
        )
        star = StarExpression()
        rel2 = rel.select(star)
        res = rel2.fetchall()
        assert res == [(1, 2)]

        # With exclude list
        star = StarExpression(exclude=['a'])
        rel2 = rel.select(star)
        res = rel2.fetchall()
        assert res == [(2,)]

    def test_struct_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                2 as b
        """
        )

        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        expr = FunctionExpression('struct_pack', col1, col2).alias('struct')

        rel = rel.select(expr)
        res = rel.fetchall()
        assert res == [({'a': 1, 'b': 2},)]

    def test_function_expression_udf(self):
        con = duckdb.connect()

        def my_simple_func(a: int, b: int, c: int) -> int:
            return a + b + c

        con.create_function('my_func', my_simple_func)

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c
        """
        )
        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        col3 = ColumnExpression('c')
        expr = FunctionExpression('my_func', col1, col2, col3)
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(6,)]

    def test_function_expression_basic(self):
        con = duckdb.connect()

        rel = con.sql(
            """
                FROM (VALUES
                    (1, 'test', 3),
                    (2, 'this is a long string', 7),
                    (3, 'medium length', 4)
                ) tbl(text, start, "end")
            """
        )
        expr = FunctionExpression('array_slice', "start", "text", "end")
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [('tes',), ('his is',), ('di',)]

    def test_column_expression_function_coverage(self):
        con = duckdb.connect()

        con.execute(
            """
            CREATE TABLE tbl as FROM (
                VALUES
                    ('a', 'b', 'c'),
                    ('d', 'e', 'f'),
                    ('g', 'h', 'i')
            ) t(c0, c1, c2)
        """
        )

        rel = con.table('tbl')
        expr = FunctionExpression('||', FunctionExpression('||', 'c0', 'c1'), 'c2')
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [('abc',), ('def',), ('ghi',)]

    def test_function_expression_aggregate(self):
        con = duckdb.connect()

        rel = con.sql(
            """
                FROM (VALUES
                    ('test'),
                    ('this is a long string'),
                    ('medium length'),
                ) tbl(text)
            """
        )
        expr = FunctionExpression('first', 'text')
        with pytest.raises(
            duckdb.BinderException, match='Binder Error: Aggregates cannot be present in a Project relation!'
        ):
            rel2 = rel.select(expr)

    def test_case_expression(self):
        con = duckdb.connect()

        rel = con.sql(
            """
            select
                1 as a,
                2 as b,
                3 as c,
        """
        )

        col1 = ColumnExpression('a')
        col2 = ColumnExpression('b')
        col3 = ColumnExpression('c')

        const1 = ConstantExpression(IntegerValue(1))
        # CASE WHEN col1 > 1 THEN 5 ELSE NULL
        case1 = CaseExpression(col1 > const1, ConstantExpression(IntegerValue(5)))

        rel2 = rel.select(case1)
        res = rel2.fetchall()
        assert res == [(None,)]

        # CASE WHEN col1 > 1 THEN 5 WHEN col2 < col1 THEN 10 ELSE NULL
        case2 = case1.when(col2 < col1, ConstantExpression(IntegerValue(10)))
        rel2 = rel.select(case2)
        res = rel2.fetchall()
        assert res == [(None,)]

        # CASE WHEN col1 > 1 THEN 5 WHEN col2 < col1 THEN 10 ELSE 42
        case3 = case2.otherwise(ConstantExpression(IntegerValue(42)))
        rel2 = rel.select(case3)
        res = rel2.fetchall()
        assert res == [(42,)]

        # CASE WHEN col3 = col3 THEN 21 WHEN col3 > col1 THEN col3 ELSE col2
        case4 = (
            CaseExpression(col3 == col3, ConstantExpression(IntegerValue(21))).when(col3 > col1, col3).otherwise(col2)
        )
        rel2 = rel.select(case4)
        res = rel2.fetchall()
        assert res == [(21,)]

    def test_cast_expression(self):
        con = duckdb.connect()

        rel = con.sql("select '2022/01/21' as a")
        assert rel.types == [VARCHAR]

        col = ColumnExpression("a").cast(TIMESTAMP)
        rel = rel.select(col)
        assert rel.types == [TIMESTAMP]

        res = rel.fetchall()
        assert res == [(datetime.datetime(2022, 1, 21, 0, 0),)]

    def test_implicit_constant_conversion(self):
        con = duckdb.connect()
        rel = con.sql("select 42")
        res = rel.select(5).fetchall()
        assert res == [(5,)]

    def test_numeric_overflow(self):
        con = duckdb.connect()
        rel = con.sql('select 3000::SHORT salary')
        with pytest.raises(duckdb.OutOfRangeException, match="Overflow in multiplication of INT16"):
            expr = ColumnExpression("salary") * 100
            rel2 = rel.select(expr)
            res = rel2.fetchall()

        with pytest.raises(duckdb.OutOfRangeException, match="Overflow in multiplication of INT16"):
            val = duckdb.Value(100, duckdb.typing.TINYINT)
            expr = ColumnExpression("salary") * val
            rel2 = rel.select(expr)
            res = rel2.fetchall()

    def test_struct_column_expression(self):
        con = duckdb.connect()
        rel = con.sql("select {'l': 1, 'ee': 33, 't': 7} as leet")
        expr = ColumnExpression("leet.ee")
        rel2 = rel.select(expr)
        res = rel2.fetchall()
        assert res == [(33,)]

    def test_filter_equality(self, filter_rel):
        assert len(filter_rel.fetchall()) == 5

        expr = ColumnExpression("a") == 1
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 2
        assert res == [(1, 'a'), (1, 'b')]

    def test_filter_not(self, filter_rel):
        expr = ColumnExpression("a") == 1
        # NOT operator
        expr = ~expr
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 3
        assert res == [(2, 'b'), (3, 'c'), (4, 'a')]

    def test_filter_and(self, filter_rel):
        expr = ColumnExpression("a") == 1
        expr = ~expr
        # AND operator

        expr = expr & ('b' != ConstantExpression('b'))
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 2
        assert res == [(3, 'c'), (4, 'a')]

    def test_filter_or(self, filter_rel):
        # OR operator
        expr = (ColumnExpression("a") == 1) | (ColumnExpression("a") == 4)
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 3
        assert res == [(1, 'a'), (1, 'b'), (4, 'a')]

    def test_filter_mixed(self, filter_rel):
        # Mixed
        expr = (ColumnExpression("b") == ConstantExpression("a")) & (
            (ColumnExpression("a") == 1) | (ColumnExpression("a") == 4)
        )
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 2
        assert res == [(1, 'a'), (4, 'a')]

    def test_filter_in(self, filter_rel):
        # IN expression
        expr = ColumnExpression("a")
        expr = expr.isin(1, 2)
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 3
        assert res == [(1, 'a'), (1, 'b'), (2, 'b')]

    def test_filter_not_in(self, filter_rel):
        expr = ColumnExpression("a")
        expr = expr.isin(1, 2)
        # NOT IN expression
        expr = ~expr
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 2
        assert res == [(3, 'c'), (4, 'a')]

        # NOT IN expression
        expr = ColumnExpression("a")
        expr = expr.isnotin(1, 2)
        rel2 = filter_rel.filter(expr)
        res = rel2.fetchall()
        assert len(res) == 2
        assert res == [(3, 'c'), (4, 'a')]

    def test_null(self):
        con = duckdb.connect()
        rel = con.sql(
            """
            select * from (VALUES
                (1, 'a'),
                (2, 'b'),
                (3, NULL),
                (4, 'c'),
                (5, 'a')
            ) tbl(a, b)
        """
        )

        b = ColumnExpression("b")

        res = rel.select(b.isnull()).fetchall()
        assert res == [(False,), (False,), (True,), (False,), (False,)]

        res2 = rel.filter(b.isnotnull()).fetchall()
        assert res2 == [(1, 'a'), (2, 'b'), (4, 'c'), (5, 'a')]

    def test_sort(self):
        con = duckdb.connect()
        rel = con.sql(
            """
            select * from (VALUES
                (1, 'a'),
                (2, 'b'),
                (3, NULL),
                (4, 'c'),
                (5, 'a')
            ) tbl(a, b)
        """
        )
        # Ascending sort order

        a = ColumnExpression("a")
        b = ColumnExpression("b")

        rel2 = rel.sort(a.asc())
        res = rel2.a.fetchall()
        assert res == [(1,), (2,), (3,), (4,), (5,)]

        # Descending sort order
        rel2 = rel.sort(a.desc())
        res = rel2.a.fetchall()
        assert res == [(5,), (4,), (3,), (2,), (1,)]

        # Nulls first
        rel2 = rel.sort(b.desc().nulls_first())
        res = rel2.b.fetchall()
        assert res == [(None,), ('c',), ('b',), ('a',), ('a',)]

        # Nulls last
        rel2 = rel.sort(b.desc().nulls_last())
        res = rel2.b.fetchall()
        assert res == [('c',), ('b',), ('a',), ('a',), (None,)]



import duckdb
import os

try:
    import pyarrow
    import pyarrow.parquet
    import pyarrow.dataset
    import numpy as np

    can_run = True
except:
    can_run = False


class TestArrowRecordBatchReader(object):
    def test_parallel_reader(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        batches = [r for r in userdata_parquet_dataset.to_batches()]
        reader = pyarrow.dataset.Scanner.from_batches(batches, schema=userdata_parquet_dataset.schema).to_reader()

        rel = duckdb_conn.from_arrow(reader)

        assert (
            rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)').execute().fetchone()[0] == 12
        )
        # The reader is already consumed so this should be 0
        assert (
            rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)').execute().fetchone()[0] == 0
        )

    def test_parallel_reader_replacement_scans(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        batches = [r for r in userdata_parquet_dataset.to_batches()]
        reader = pyarrow.dataset.Scanner.from_batches(batches, schema=userdata_parquet_dataset.schema).to_reader()

        assert (
            duckdb_conn.execute(
                "select count(*) from reader where first_name=\'Jose\' and salary > 134708.82"
            ).fetchone()[0]
            == 12
        )
        assert (
            duckdb_conn.execute(
                "select count(*) from reader where first_name=\'Jose\' and salary > 134708.82"
            ).fetchone()[0]
            == 0
        )

    def test_parallel_reader_register(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        batches = [r for r in userdata_parquet_dataset.to_batches()]
        reader = pyarrow.dataset.Scanner.from_batches(batches, schema=userdata_parquet_dataset.schema).to_reader()

        duckdb_conn.register("bla", reader)

        assert (
            duckdb_conn.execute("select count(*) from bla where first_name=\'Jose\' and salary > 134708.82").fetchone()[
                0
            ]
            == 12
        )
        assert (
            duckdb_conn.execute("select count(*) from bla where first_name=\'Jose\' and salary > 134708.82").fetchone()[
                0
            ]
            == 0
        )

    def test_parallel_reader_default_conn(self, duckdb_cursor):
        if not can_run:
            return

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        batches = [r for r in userdata_parquet_dataset.to_batches()]
        reader = pyarrow.dataset.Scanner.from_batches(batches, schema=userdata_parquet_dataset.schema).to_reader()

        rel = duckdb.from_arrow(reader)

        assert (
            rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)').execute().fetchone()[0] == 12
        )
        # The reader is already consumed so this should be 0
        assert (
            rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)').execute().fetchone()[0] == 0
        )



import duckdb
import tempfile
import os
import pandas as pd


class TestRelationApi(object):
    def test_readonly(self, duckdb_cursor):
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3], "j": ["one", "two", "three"]})

        def test_rel(rel, duckdb_cursor):
            res = (
                rel.filter('i < 3')
                .order('j')
                .project('i')
                .union(rel.filter('i > 2').project('i'))
                .join(rel.set_alias('a1'), 'i')
                .project('CAST(i as BIGINT) i, j')
                .order('i')
            )
            pd.testing.assert_frame_equal(res.to_df(), test_df)
            res3 = duckdb_cursor.from_df(res.to_df()).to_df()
            pd.testing.assert_frame_equal(res3, test_df)

            df_sql = res.query('x', 'select CAST(i as BIGINT) i, j from x')
            pd.testing.assert_frame_equal(df_sql.df(), test_df)

            res2 = res.aggregate('i, count(j) as cj', 'i').order('i')
            cmp_df = pd.DataFrame.from_dict({"i": [1, 2, 3], "cj": [1, 1, 1]})
            pd.testing.assert_frame_equal(res2.to_df(), cmp_df)

            duckdb_cursor.execute('DROP TABLE IF EXISTS a2')
            rel.create('a2')
            rel_a2 = duckdb_cursor.table('a2').project('CAST(i as BIGINT) i, j').to_df()
            pd.testing.assert_frame_equal(rel_a2, test_df)

            duckdb_cursor.execute('DROP TABLE IF EXISTS a3')
            duckdb_cursor.execute('CREATE TABLE a3 (i INTEGER, j STRING)')
            rel.insert_into('a3')
            rel_a3 = duckdb_cursor.table('a3').project('CAST(i as BIGINT) i, j').to_df()
            pd.testing.assert_frame_equal(rel_a3, test_df)

        duckdb_cursor.execute('CREATE TABLE a (i INTEGER, j STRING)')
        duckdb_cursor.execute("INSERT INTO a VALUES (1, 'one'), (2, 'two'), (3, 'three')")
        duckdb_cursor.execute('CREATE VIEW v AS SELECT * FROM a')

        duckdb_cursor.execute('CREATE TEMPORARY TABLE at (i INTEGER)')
        duckdb_cursor.execute('CREATE TEMPORARY VIEW vt AS SELECT * FROM at')

        rel_a = duckdb_cursor.table('a')
        rel_v = duckdb_cursor.view('v')
        # rel_at = duckdb_cursor.table('at')
        # rel_vt = duckdb_cursor.view('vt')

        rel_df = duckdb_cursor.from_df(test_df)

        test_rel(rel_a, duckdb_cursor)
        test_rel(rel_v, duckdb_cursor)
        test_rel(rel_df, duckdb_cursor)

    def test_fromquery(self, duckdb_cursor):
        assert duckdb.from_query('select 42').fetchone()[0] == 42
        assert duckdb_cursor.query('select 43').fetchone()[0] == 43

        # assert duckdb_cursor.from_query('select 44').execute().fetchone()[0] == 44
        # assert duckdb_cursor.from_query('select 45').execute().fetchone()[0] == 45
        # assert duckdb_cursor.from_query('select 45').execute().fetchone()[0] == 45


# cursor = duckdb.connect().cursor()
# TestRelationApi().test_readonly(cursor)



import duckdb
import pytest
import os
import tempfile
import pandas as pd

VARCHAR = duckdb.typing.VARCHAR
BIGINT = duckdb.typing.BIGINT

filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'binary_string.parquet')


@pytest.fixture(scope="session")
def tmp_parquets(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp('parquets', numbered=True)
    tmp_parquets = [str(tmp_dir / ('tmp' + str(i) + '.parquet')) for i in range(1, 4)]
    return tmp_parquets


class TestParquet(object):
    def test_scan_binary(self, duckdb_cursor):
        conn = duckdb.connect()
        res = conn.execute("SELECT typeof(#1) FROM parquet_scan('" + filename + "') limit 1").fetchall()
        assert res[0] == ('BLOB',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "')").fetchall()
        assert res[0] == (b'foo',)

    def test_from_parquet_binary(self, duckdb_cursor):
        rel = duckdb.from_parquet(filename)
        assert rel.types == ['BLOB']

        res = rel.execute().fetchall()
        assert res[0] == (b'foo',)

    def test_scan_binary_as_string(self, duckdb_cursor):
        conn = duckdb.connect()
        res = conn.execute(
            "SELECT typeof(#1) FROM parquet_scan('" + filename + "',binary_as_string=True) limit 1"
        ).fetchall()
        assert res[0] == ('VARCHAR',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "',binary_as_string=True)").fetchall()
        assert res[0] == ('foo',)

    def test_from_parquet_binary_as_string(self, duckdb_cursor):
        rel = duckdb.from_parquet(filename, True)
        assert rel.types == [VARCHAR]

        res = rel.execute().fetchall()
        assert res[0] == ('foo',)

    def test_from_parquet_file_row_number(self, duckdb_cursor):
        rel = duckdb.from_parquet(filename, binary_as_string=True, file_row_number=True)
        assert rel.types == [VARCHAR, BIGINT]

        res = rel.execute().fetchall()
        assert res[0] == (
            'foo',
            0,
        )

    def test_from_parquet_filename(self, duckdb_cursor):
        rel = duckdb.from_parquet(filename, binary_as_string=True, filename=True)
        assert rel.types == [VARCHAR, VARCHAR]

        res = rel.execute().fetchall()
        assert res[0] == (
            'foo',
            filename,
        )

    def test_from_parquet_list_binary_as_string(self, duckdb_cursor):
        rel = duckdb.from_parquet([filename], binary_as_string=True)
        assert rel.types == [VARCHAR]

        res = rel.execute().fetchall()
        assert res[0] == ('foo',)

    def test_from_parquet_list_file_row_number(self, duckdb_cursor):
        rel = duckdb.from_parquet([filename], binary_as_string=True, file_row_number=True)
        assert rel.types == [VARCHAR, BIGINT]

        res = rel.execute().fetchall()
        assert res[0] == (
            'foo',
            0,
        )

    def test_from_parquet_list_filename(self, duckdb_cursor):
        rel = duckdb.from_parquet([filename], binary_as_string=True, filename=True)
        assert rel.types == [VARCHAR, VARCHAR]

        res = rel.execute().fetchall()
        assert res[0] == (
            'foo',
            filename,
        )

    def test_parquet_binary_as_string_pragma(self, duckdb_cursor):
        conn = duckdb.connect()
        res = conn.execute("SELECT typeof(#1) FROM parquet_scan('" + filename + "') limit 1").fetchall()
        assert res[0] == ('BLOB',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "')").fetchall()
        assert res[0] == (b'foo',)

        conn.execute("PRAGMA binary_as_string=1")

        res = conn.execute("SELECT typeof(#1) FROM parquet_scan('" + filename + "') limit 1").fetchall()
        assert res[0] == ('VARCHAR',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "')").fetchall()
        assert res[0] == ('foo',)

        res = conn.execute(
            "SELECT typeof(#1) FROM parquet_scan('" + filename + "',binary_as_string=False) limit 1"
        ).fetchall()
        assert res[0] == ('BLOB',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "',binary_as_string=False)").fetchall()
        assert res[0] == (b'foo',)

        conn.execute("PRAGMA binary_as_string=0")

        res = conn.execute("SELECT typeof(#1) FROM parquet_scan('" + filename + "') limit 1").fetchall()
        assert res[0] == ('BLOB',)

        res = conn.execute("SELECT * FROM parquet_scan('" + filename + "')").fetchall()
        assert res[0] == (b'foo',)

    def test_from_parquet_binary_as_string_default_conn(self, duckdb_cursor):
        duckdb.default_connection.execute("PRAGMA binary_as_string=1")

        rel = duckdb.from_parquet(filename, True)
        assert rel.types == [VARCHAR]

        res = rel.execute().fetchall()
        assert res[0] == ('foo',)

    def test_from_parquet_union_by_name(self, tmp_parquets):
        conn = duckdb.connect()

        conn.execute(
            "copy (from (values (1::bigint), (2::bigint), (9223372036854775807::bigint)) t(a)) to '"
            + tmp_parquets[0]
            + "' (format 'parquet');"
        )

        conn.execute(
            "copy (from (values (3::integer, 4::integer), (5::integer, 6::integer)) t(a, b)) to '"
            + tmp_parquets[1]
            + "' (format 'parquet');"
        )

        conn.execute(
            "copy (from (values (100::integer, 101::integer), (102::integer, 103::integer)) t(a, c)) to '"
            + tmp_parquets[2]
            + "' (format 'parquet');"
        )

        rel = duckdb.from_parquet(tmp_parquets, union_by_name=True).order('a')
        assert rel.execute().fetchall() == [
            (
                1,
                None,
                None,
            ),
            (
                2,
                None,
                None,
            ),
            (
                3,
                4,
                None,
            ),
            (
                5,
                6,
                None,
            ),
            (
                100,
                None,
                101,
            ),
            (
                102,
                None,
                103,
            ),
            (
                9223372036854775807,
                None,
                None,
            ),
        ]



import duckdb
import pytest
import pandas as pd
import duckdb

try:
    import pyarrow as pa

    can_run = True
except:
    can_run = False


class TestArrowBatchIndex(object):
    def test_arrow_batch_index(self, duckdb_cursor):
        if not can_run:
            return
        con = duckdb.connect()
        df = con.execute('SELECT * FROM range(10000000) t(i)').df()
        arrow_tbl = pa.Table.from_pandas(df)

        con.execute('CREATE TABLE tbl AS SELECT * FROM arrow_tbl')

        result = con.execute('SELECT * FROM tbl LIMIT 5').fetchall()
        assert [x[0] for x in result] == [0, 1, 2, 3, 4]

        result = con.execute('SELECT * FROM tbl LIMIT 5 OFFSET 777778').fetchall()
        assert [x[0] for x in result] == [777778, 777779, 777780, 777781, 777782]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ctypes
import os
import platform
import subprocess
import sys
import traceback
from functools import lru_cache
from glob import glob
from os.path import exists
from typing import TextIO

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext


class CompilerLauncherMixin:
    """Add "compiler launchers" to distutils.

    We use this to be able to run the build using "ccache".

    A compiler launcher is a program that is invoked instead of invoking the
    compiler directly. It is passed the full compiler invocation command line.

    A similar feature exists in CMake, see
    https://cmake.org/cmake/help/latest/prop_tgt/LANG_COMPILER_LAUNCHER.html.
    """

    __is_set_up = False

    def build_extensions(self):
        # Integrate into "build_ext"
        self.__setup()
        super().build_extensions()

    def build_libraries(self):
        # Integrate into "build_clib"
        self.__setup()
        super().build_extensions()

    def __setup(self):
        if self.__is_set_up:
            return
        self.__is_set_up = True
        compiler_launcher = os.getenv("DISTUTILS_C_COMPILER_LAUNCHER")
        if compiler_launcher:

            def spawn_with_compiler_launcher(cmd, **kwargs):
                if platform.system() == 'Windows' and len(' '.join(cmd)) > 32766:
                    raise Exception("command too long: " + ' '.join(cmd))

                exclude_programs = ("link.exe",)
                if not cmd[0].endswith(exclude_programs):
                    cmd = [compiler_launcher] + cmd
                try:
                    return original_spawn(cmd, **kwargs)
                except Exception:
                    traceback.print_exc()
                    raise

            original_spawn = self.compiler.spawn
            self.compiler.spawn = spawn_with_compiler_launcher

            # for some reason, the copy of distutils included with setuptools  (instead of the one that was previously
            # part of the stdlib) pushes the link call over the maximum command line length on windows.
            # We can fix this by using special windows short paths, of the form C:\Program ~1\...

            def link_with_short_paths(
                target_desc,
                objects,
                *args,
                **kwargs,
            ):
                return original_link(target_desc, [get_short_path(x) for x in objects], *args, **kwargs)

            original_link = self.compiler.link
            self.compiler.link = link_with_short_paths


def get_short_path(long_name: str) -> str:
    """
    Gets the short path name of a given long path.
    http://stackoverflow.com/a/23598461/200291
    """

    @lru_cache()
    def get_short_path_name_w():
        from ctypes import wintypes

        kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
        _GetShortPathNameW = kernel32.GetShortPathNameW
        _GetShortPathNameW.argtypes = [wintypes.LPCWSTR, wintypes.LPWSTR, wintypes.DWORD]
        _GetShortPathNameW.restype = wintypes.DWORD
        return _GetShortPathNameW

    if platform.system() != 'Windows':
        return long_name

    assert exists(long_name), long_name
    gspn = get_short_path_name_w()

    output_buf_size = 0
    while True:
        output_buf = ctypes.create_unicode_buffer(output_buf_size)
        needed = gspn(long_name, output_buf, output_buf_size)
        if output_buf_size >= needed:
            return output_buf.value or long_name
        else:
            output_buf_size = needed


class build_ext(CompilerLauncherMixin, _build_ext):
    pass


lib_name = 'duckdb'

extensions = ['parquet', 'icu', 'fts', 'tpch', 'tpcds', 'json']

if platform.system() == 'Windows':
    extensions = ['parquet', 'icu', 'fts', 'tpch', 'json']

is_android = hasattr(sys, 'getandroidapilevel')
use_jemalloc = (
    not is_android
    and platform.system() == 'Linux'
    and platform.architecture()[0] == '64bit'
    and platform.machine() == 'x86_64'
)

if use_jemalloc:
    extensions.append('jemalloc')

unity_build = 0
if 'DUCKDB_BUILD_UNITY' in os.environ:
    unity_build = 16

try:
    import pybind11
except ImportError:
    raise Exception(
        'pybind11 could not be imported. This usually means you\'re calling setup.py directly, or using a version of pip that doesn\'t support PEP517'
    ) from None

# speed up compilation with: -j = cpu_number() on non Windows machines
if os.name != 'nt' and os.environ.get('DUCKDB_DISABLE_PARALLEL_COMPILE', '') != '1':
    from pybind11.setup_helpers import ParallelCompile

    ParallelCompile().install()


def open_utf8(fpath, flags):
    import sys

    if sys.version_info[0] < 3:
        return open(fpath, flags)
    else:
        return open(fpath, flags, encoding="utf8")


# make sure we are in the right directory
os.chdir(os.path.dirname(os.path.realpath(__file__)))

if os.name == 'nt':
    # windows:
    toolchain_args = ['/wd4244', '/wd4267', '/wd4200', '/wd26451', '/wd26495', '/D_CRT_SECURE_NO_WARNINGS', '/utf-8']
else:
    # macos/linux
    toolchain_args = ['-std=c++11', '-g0']
    if 'DUCKDEBUG' in os.environ:
        toolchain_args = ['-std=c++11', '-Wall', '-O0', '-g']
if 'DUCKDB_INSTALL_USER' in os.environ and 'install' in sys.argv:
    sys.argv.append('--user')

existing_duckdb_dir = ''
libraries = []
if 'DUCKDB_BINARY_DIR' in os.environ:
    existing_duckdb_dir = os.environ['DUCKDB_BINARY_DIR']
if 'DUCKDB_COMPILE_FLAGS' in os.environ:
    toolchain_args = ['-std=c++11'] + os.environ['DUCKDB_COMPILE_FLAGS'].split()
if 'DUCKDB_LIBS' in os.environ:
    libraries = os.environ['DUCKDB_LIBS'].split(' ')

define_macros = [('DUCKDB_PYTHON_LIB_NAME', lib_name)]

if platform.system() == 'Darwin':
    toolchain_args.extend(['-stdlib=libc++', '-mmacosx-version-min=10.7'])

if platform.system() == 'Windows':
    define_macros.extend([('DUCKDB_BUILD_LIBRARY', None), ('WIN32', None)])

if 'BUILD_HTTPFS' in os.environ:
    libraries += ['crypto', 'ssl']
    extensions += ['httpfs']

for ext in extensions:
    define_macros.append(('DUCKDB_EXTENSION_{}_LINKED'.format(ext.upper()), None))

define_macros.extend([('DUCKDB_EXTENSION_AUTOLOAD_DEFAULT', '1'), ('DUCKDB_EXTENSION_AUTOINSTALL_DEFAULT', '1')])

linker_args = toolchain_args[:]
if platform.system() == 'Windows':
    linker_args.extend(['rstrtmgr.lib', 'bcrypt.lib'])

short_paths = False
if platform.system() == 'Windows':
    short_paths = True

extra_files = []
header_files = []


def list_source_files(directory):
    sources = glob('src/**/*.cpp', recursive=True)
    return sources


script_path = os.path.dirname(os.path.abspath(__file__))
main_include_path = os.path.join(script_path, 'src', 'include')
main_source_path = os.path.join(script_path, 'src')
main_source_files = ['duckdb_python.cpp'] + list_source_files(main_source_path)

include_directories = [main_include_path, pybind11.get_include(False), pybind11.get_include(True)]
if 'BUILD_HTTPFS' in os.environ and 'OPENSSL_ROOT_DIR' in os.environ:
    include_directories += [os.path.join(os.environ['OPENSSL_ROOT_DIR'], 'include')]


def exclude_extensions(f: TextIO):
    files = [x for x in f.read().split('\n') if len(x) > 0]
    if use_jemalloc:
        return files
    else:
        return [x for x in files if 'jemalloc' not in x]


if len(existing_duckdb_dir) == 0:
    # no existing library supplied: compile everything from source
    source_files = main_source_files

    # check if amalgamation exists
    if os.path.isfile(os.path.join(script_path, '..', '..', 'scripts', 'amalgamation.py')):
        # amalgamation exists: compiling from source directory
        # copy all source files to the current directory
        sys.path.append(os.path.join(script_path, '..', '..', 'scripts'))
        import package_build

        (source_list, include_list, original_sources) = package_build.build_package(
            os.path.join(script_path, "duckdb_build"), extensions, False, unity_build, "duckdb_build", short_paths
        )

        duckdb_sources = [
            os.path.sep.join(package_build.get_relative_path(script_path, x).split('/')) for x in source_list
        ]
        duckdb_sources.sort()

        original_sources = [os.path.join("duckdb_build", x) for x in original_sources]

        duckdb_includes = [os.path.join("duckdb_build", x) for x in include_list]
        duckdb_includes += ["duckdb_build"]

        # gather the include files
        import amalgamation

        header_files = amalgamation.list_includes_files(duckdb_includes)

        # write the source list, include list and git hash to separate files
        with open_utf8('sources.list', 'w+') as f:
            for source_file in duckdb_sources:
                f.write(source_file + "\n")

        with open_utf8('includes.list', 'w+') as f:
            for include_file in duckdb_includes:
                f.write(include_file + '\n')

        extra_files = ['sources.list', 'includes.list'] + original_sources
    else:
        # if amalgamation does not exist, we are in a package distribution
        # read the include files, source list and include files from the supplied lists
        with open_utf8('sources.list', 'r') as f:
            duckdb_sources = exclude_extensions(f)

        with open_utf8('includes.list', 'r') as f:
            duckdb_includes = exclude_extensions(f)

    source_files += duckdb_sources
    include_directories = duckdb_includes + include_directories

    libduckdb = Extension(
        lib_name + '.duckdb',
        include_dirs=include_directories,
        sources=source_files,
        extra_compile_args=toolchain_args,
        extra_link_args=linker_args,
        libraries=libraries,
        language='c++',
        define_macros=define_macros,
    )
else:
    sys.path.append(os.path.join(script_path, '..', '..', 'scripts'))
    import package_build

    include_directories += [os.path.join('..', '..', include) for include in package_build.third_party_includes()]
    toolchain_args += ['-I' + x for x in package_build.includes(extensions)]

    result_libraries = package_build.get_libraries(existing_duckdb_dir, libraries, extensions)
    library_dirs = [x[0] for x in result_libraries if x[0] is not None]
    libnames = [x[1] for x in result_libraries if x[1] is not None]

    libduckdb = Extension(
        lib_name + '.duckdb',
        include_dirs=include_directories,
        sources=main_source_files,
        extra_compile_args=toolchain_args,
        extra_link_args=linker_args,
        libraries=libnames,
        library_dirs=library_dirs,
        language='c++',
        define_macros=define_macros,
    )


# data files need to be formatted as [(directory, [files...]), (directory2, [files...])]
# no clue why the setup script can't do this automatically, but hey
def setup_data_files(data_files):
    directory_map = {}
    for data_file in data_files:
        normalized_fpath = os.path.sep.join(data_file.split('/'))
        splits = normalized_fpath.rsplit(os.path.sep, 1)
        if len(splits) == 1:
            # no directory specified
            directory = ""
            fname = normalized_fpath
        else:
            directory = splits[0]
            fname = splits[1]
        if directory not in directory_map:
            directory_map[directory] = []
        directory_map[directory].append(normalized_fpath)
    new_data_files = []
    for kv in directory_map.keys():
        new_data_files.append((kv, directory_map[kv]))
    return new_data_files


data_files = setup_data_files(extra_files + header_files)

packages = [
    lib_name,
    'duckdb.typing',
    'duckdb.query_graph',
    'duckdb.functional',
    'duckdb.value',
    'duckdb-stubs',
    'duckdb-stubs.functional',
    'duckdb-stubs.typing',
    'duckdb-stubs.value',
    'duckdb-stubs.value.constant',
    'adbc_driver_duckdb',
]

spark_packages = [
    'duckdb.experimental',
    'duckdb.experimental.spark',
    'duckdb.experimental.spark.sql',
    'duckdb.experimental.spark.errors',
    'duckdb.experimental.spark.errors.exceptions',
]

packages.extend(spark_packages)


# Duplicated from scripts/package_build.py
def get_git_describe():
    override_git_describe = os.getenv('OVERRIDE_GIT_DESCRIBE') or ''
    # empty override_git_describe, either since env was empty string or not existing
    # -> ask git (that can fail, so except in place)
    if len(override_git_describe) == 0:
        try:
            return subprocess.check_output(['git', 'describe', '--tags', '--long']).strip().decode('utf8')
        except subprocess.CalledProcessError:
            return "v0.0.0-0-gdeadbeeff"
    if len(override_git_describe.split('-')) == 3:
        return override_git_describe
    if len(override_git_describe.split('-')) == 1:
        override_git_describe += "-0"
    assert len(override_git_describe.split('-')) == 2
    try:
        return (
            override_git_describe
            + "-g"
            + subprocess.check_output(['git', 'log', '-1', '--format=%h']).strip().decode('utf8')
        )
    except subprocess.CalledProcessError:
        return override_git_describe + "-g" + "deadbeeff"


def get_version():
    git_describe = get_git_describe()
    version = git_describe.split('-')[0]
    return version


setup(
    name=lib_name,
    version=get_version(),
    description='DuckDB in-process database',
    keywords='DuckDB Database SQL OLAP',
    url="https://www.duckdb.org",
    long_description='See here for an introduction: https://duckdb.org/docs/api/python/overview',
    license='MIT',
    data_files=data_files,
    # NOTE: might need to be find_packages() ?
    packages=packages,
    include_package_data=True,
    python_requires='>=3.7.0',
    tests_require=['google-cloud-storage', 'mypy', 'pytest'],
    classifiers=[
        'Topic :: Database :: Database Engines/Servers',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
    ],
    ext_modules=[libduckdb],
    maintainer="Hannes Muehleisen",
    maintainer_email="hannes@cwi.nl",
    cmdclass={"build_ext": build_ext},
    project_urls={
        "Documentation": "https://duckdb.org/docs/api/python/overview",
        "Source": "https://github.com/duckdb/duckdb/blob/main/tools/pythonpkg",
        "Issues": "https://github.com/duckdb/duckdb/issues",
        "Changelog": "https://github.com/duckdb/duckdb/releases",
    },
)



import duckdb
import os

try:
    import pyarrow
    import pyarrow.parquet

    can_run = True
except:
    can_run = False


class TestArrowView(object):
    def test_arrow_view(self, duckdb_cursor):
        if not can_run:
            return
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        cols = 'id, first_name, last_name, email, gender, ip_address, cc, country, birthdate, salary, title, comments'
        duckdb_conn = duckdb.connect()
        userdata_parquet_table = pyarrow.parquet.read_table(parquet_filename)
        userdata_parquet_table.validate(full=True)
        duckdb_conn.from_arrow(userdata_parquet_table).create_view('arrow_view')
        assert duckdb_conn.execute("PRAGMA show_tables").fetchone() == ('arrow_view',)
        assert duckdb_conn.execute("select avg(salary)::INT from arrow_view").fetchone()[0] == 149005



# cursor description

import duckdb
import tempfile
import os


def check_exception(f):
    had_exception = False
    try:
        f()
    except:
        had_exception = True
    assert had_exception


class TestReadOnly(object):
    def test_readonly(self, duckdb_cursor):
        fd, db = tempfile.mkstemp()
        os.close(fd)
        os.remove(db)

        # this is forbidden
        check_exception(lambda: duckdb.connect(":memory:", True))

        con_rw = duckdb.connect(db, False)
        con_rw.cursor().execute("create table a (i integer)")
        con_rw.cursor().execute("insert into a values (42)")
        con_rw.close()

        con_ro = duckdb.connect(db, True)
        con_ro.cursor().execute("select * from a").fetchall()
        check_exception(lambda: con_ro.execute("delete from a"))
        con_ro.close()

        con_rw = duckdb.connect(db, False)
        con_rw.cursor().execute("drop table a")
        con_rw.close()



import duckdb
from pytest import raises
from duckdb import NotImplementedException, InvalidInputException
from duckdb.value.constant import (
    Value,
    NullValue,
    BooleanValue,
    UnsignedBinaryValue,
    UnsignedShortValue,
    UnsignedIntegerValue,
    UnsignedLongValue,
    BinaryValue,
    ShortValue,
    IntegerValue,
    LongValue,
    HugeIntegerValue,
    UnsignedHugeIntegerValue,
    FloatValue,
    DoubleValue,
    DecimalValue,
    StringValue,
    UUIDValue,
    BitValue,
    BlobValue,
    DateValue,
    IntervalValue,
    TimestampValue,
    TimestampSecondValue,
    TimestampMilisecondValue,
    TimestampNanosecondValue,
    TimestampTimeZoneValue,
    TimeValue,
    TimeTimeZoneValue,
)
import uuid
import datetime
import pytest
import decimal

from duckdb.typing import (
    SQLNULL,
    BOOLEAN,
    TINYINT,
    UTINYINT,
    SMALLINT,
    USMALLINT,
    INTEGER,
    UINTEGER,
    BIGINT,
    UBIGINT,
    HUGEINT,
    UHUGEINT,
    UUID,
    FLOAT,
    DOUBLE,
    DATE,
    TIMESTAMP,
    TIMESTAMP_MS,
    TIMESTAMP_NS,
    TIMESTAMP_S,
    TIME,
    TIME_TZ,
    TIMESTAMP_TZ,
    VARCHAR,
    BLOB,
    BIT,
    INTERVAL,
)


class TestValue(object):
    # This excludes timezone aware values, as those are a pain to test
    @pytest.mark.parametrize(
        'item',
        [
            (BOOLEAN, BooleanValue(True), True),
            (UTINYINT, UnsignedBinaryValue(129), 129),
            (USMALLINT, UnsignedShortValue(12356), 12356),
            (UINTEGER, UnsignedIntegerValue(2435035), 2435035),
            (UBIGINT, UnsignedLongValue(243503523), 243503523),
            (TINYINT, BinaryValue(-1), -1),
            (SMALLINT, ShortValue(-1), -1),
            (INTEGER, IntegerValue(-1), -1),
            (BIGINT, LongValue(-1), -1),
            (HUGEINT, HugeIntegerValue(-1), -1),
            (UHUGEINT, UnsignedHugeIntegerValue(24350352345), 24350352345),
            (FLOAT, FloatValue(1.8349000215530396), 1.8349000215530396),
            (DOUBLE, DoubleValue(0.23234234234), 0.23234234234),
            (
                duckdb.decimal_type(12, 8),
                DecimalValue(decimal.Decimal('1234.12345678'), 12, 8),
                decimal.Decimal('1234.12345678'),
            ),
            (VARCHAR, StringValue('this is a long string'), 'this is a long string'),
            (
                UUID,
                UUIDValue(uuid.UUID('ffffffff-ffff-ffff-ffff-ffffffffffff')),
                uuid.UUID('ffffffff-ffff-ffff-ffff-ffffffffffff'),
            ),
            (BIT, BitValue(b'010101010101'), '010101010101'),
            (BLOB, BlobValue(b'\x00\x00\x00a'), b'\x00\x00\x00a'),
            (DATE, DateValue(datetime.date(2000, 5, 4)), datetime.date(2000, 5, 4)),
            (INTERVAL, IntervalValue(datetime.timedelta(days=5)), datetime.timedelta(days=5)),
            (
                TIMESTAMP,
                TimestampValue(datetime.datetime(1970, 3, 21, 12, 5, 43, 120)),
                datetime.datetime(1970, 3, 21, 12, 5, 43, 120),
            ),
            (SQLNULL, NullValue(), None),
            (TIME, TimeValue(datetime.time(12, 3, 12, 80)), datetime.time(12, 3, 12, 80)),
        ],
    )
    def test_value_helpers(self, item):
        expected_type = item[0]
        value_object = item[1]
        expected_value = item[2]

        con = duckdb.connect()
        observed_type = con.execute('select typeof(a) from (select $1) tbl(a)', [value_object]).fetchall()[0][0]
        assert observed_type == str(expected_type)

        con.execute('select $1', [value_object])
        result = con.fetchone()
        result = result[0]
        assert result == expected_value

    def test_float_to_decimal_prevention(self):
        value = DecimalValue(1.2345, 12, 8)

        con = duckdb.connect()
        with pytest.raises(duckdb.ConversionException, match="Can't losslessly convert"):
            con.execute('select $1', [value]).fetchall()

    @pytest.mark.parametrize(
        'value',
        [
            TimestampSecondValue(datetime.datetime(1970, 3, 21, 12, 36, 43)),
            TimestampMilisecondValue(datetime.datetime(1970, 3, 21, 12, 36, 43)),
            TimestampNanosecondValue(datetime.datetime(1970, 3, 21, 12, 36, 43)),
        ],
    )
    def test_timestamp_sec_not_supported(self, value):
        con = duckdb.connect()
        with pytest.raises(
            duckdb.NotImplementedException, match="Conversion from 'datetime' to type .* is not implemented yet"
        ):
            con.execute('select $1', [value]).fetchall()

    @pytest.mark.parametrize(
        'target_type,test_value,expected_conversion_success',
        [
            (TINYINT, 0, True),
            (TINYINT, 255, False),
            (TINYINT, -128, True),
            (UTINYINT, 80, True),
            (UTINYINT, -1, False),
            (UTINYINT, 255, True),
            (SMALLINT, 0, True),
            (SMALLINT, 128, True),
            (SMALLINT, -255, True),
            (SMALLINT, -32780, False),
            (USMALLINT, 0, True),
            (USMALLINT, -1, False),
            (USMALLINT, 1337, True),
            (USMALLINT, 32780, True),
            (INTEGER, 0, True),
            (INTEGER, 32780, True),
            (INTEGER, -32780, True),
            (INTEGER, -1337, True),
            (UINTEGER, 0, True),
            (UINTEGER, -1337, False),
            (UINTEGER, 65534, True),
            (BIGINT, 0, True),
            (BIGINT, -1234567, True),
            (BIGINT, 9223372036854775808, False),
            (UBIGINT, 9223372036854775808, True),
            (UBIGINT, -1, False),
            (UBIGINT, 18446744073709551615, True),
            (HUGEINT, -9223372036854775808, True),
            (HUGEINT, 9223372036854775807, True),
            (HUGEINT, 0, True),
            (HUGEINT, -1337, True),
            (HUGEINT, 12334214123, True),
        ],
    )
    def test_numeric_values(self, target_type, test_value, expected_conversion_success):
        value = Value(test_value, target_type)
        con = duckdb.connect()

        work = lambda: con.execute('select typeof(a) from (select $1) tbl(a)', [value]).fetchall()

        if expected_conversion_success:
            res = work()
            assert str(target_type) == res[0][0]
        else:
            with raises((NotImplementedException, InvalidInputException)):
                work()



from .column import Column, _get_expr
from typing import Any, Callable, overload, Union

from duckdb import CaseExpression, ConstantExpression, ColumnExpression, FunctionExpression, Expression
from ._typing import ColumnOrName
from ..exception import ContributionsAcceptedError


def _invoke_function_over_columns(name: str, *cols: "ColumnOrName") -> Column:
    """
    Invokes n-ary JVM function identified by name
    and wraps the result with :class:`~pyspark.sql.Column`.
    """
    cols = [_to_column(expr) for expr in cols]
    return _invoke_function(name, *cols)


def col(column: str):
    return Column(ColumnExpression(column))


def upper(col: "ColumnOrName") -> Column:
    """
    Converts a string expression to upper case.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        upper case values.

    Examples
    --------
    >>> df = spark.createDataFrame(["Spark", "PySpark", "Pandas API"], "STRING")
    >>> df.select(upper("value")).show()
    +------------+
    |upper(value)|
    +------------+
    |       SPARK|
    |     PYSPARK|
    |  PANDAS API|
    +------------+
    """
    return _invoke_function_over_columns("upper", col)


def when(condition: "Column", value: Any) -> Column:
    if not isinstance(condition, Column):
        raise TypeError("condition should be a Column")
    v = _get_expr(value)
    expr = CaseExpression(condition.expr, v)
    return Column(expr)


def _inner_expr_or_val(val):
    return val.expr if isinstance(val, Column) else val


def struct(*cols: Column) -> Column:
    return Column(FunctionExpression('struct_pack', *[_inner_expr_or_val(x) for x in cols]))


def lit(col: Any) -> Column:
    return col if isinstance(col, Column) else Column(ConstantExpression(col))


def _invoke_function(function: str, *arguments):
    return Column(FunctionExpression(function, *arguments))


def _to_column(col: ColumnOrName) -> Expression:
    return col.expr if isinstance(col, Column) else ColumnExpression(col)


def regexp_replace(str: "ColumnOrName", pattern: str, replacement: str) -> Column:
    r"""Replace all substrings of the specified string value that match regexp with rep.

    .. versionadded:: 1.5.0

    Examples
    --------
    >>> df = spark.createDataFrame([('100-200',)], ['str'])
    >>> df.select(regexp_replace('str', r'(\d+)', '--').alias('d')).collect()
    [Row(d='-----')]
    """
    return _invoke_function(
        "regexp_replace", _to_column(str), ConstantExpression(pattern), ConstantExpression(replacement), ConstantExpression('g')
    )


def array_contains(col: "ColumnOrName", value: Any) -> Column:
    """
    Collection function: returns null if the array is null, true if the array contains the
    given value, and false otherwise.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        name of column containing array
    value :
        value or column to check for in array

    Returns
    -------
    :class:`~pyspark.sql.Column`
        a column of Boolean type.

    Examples
    --------
    >>> df = spark.createDataFrame([(["a", "b", "c"],), ([],)], ['data'])
    >>> df.select(array_contains(df.data, "a")).collect()
    [Row(array_contains(data, a)=True), Row(array_contains(data, a)=False)]
    >>> df.select(array_contains(df.data, lit("a"))).collect()
    [Row(array_contains(data, a)=True), Row(array_contains(data, a)=False)]
    """
    value = _get_expr(value)
    return _invoke_function("array_contains", _to_column(col), value)


def avg(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the average of the values in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(avg(col("id"))).show()
    +-------+
    |avg(id)|
    +-------+
    |    4.5|
    +-------+
    """
    return _invoke_function_over_columns("avg", col)


def sum(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the sum of all values in the expression.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(sum(df["id"])).show()
    +-------+
    |sum(id)|
    +-------+
    |     45|
    +-------+
    """
    return _invoke_function_over_columns("sum", col)


def max(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the maximum value of the expression in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(max(col("id"))).show()
    +-------+
    |max(id)|
    +-------+
    |      9|
    +-------+
    """
    return _invoke_function_over_columns("max", col)


def mean(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the average of the values in a group.
    An alias of :func:`avg`.

    .. versionadded:: 1.4.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(mean(df.id)).show()
    +-------+
    |avg(id)|
    +-------+
    |    4.5|
    +-------+
    """
    return _invoke_function_over_columns("mean", col)


def min(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the minimum value of the expression in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(min(df.id)).show()
    +-------+
    |min(id)|
    +-------+
    |      0|
    +-------+
    """
    return _invoke_function_over_columns("min", col)


def count(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the number of items in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    Count by all columns (start), and by a column that does not count ``None``.

    >>> df = spark.createDataFrame([(None,), ("a",), ("b",), ("c",)], schema=["alphabets"])
    >>> df.select(count(expr("*")), count(df.alphabets)).show()
    +--------+----------------+
    |count(1)|count(alphabets)|
    +--------+----------------+
    |       4|               3|
    +--------+----------------+
    """
    return _invoke_function_over_columns("count", col)


@overload
def transform(col: "ColumnOrName", f: Callable[[Column], Column]) -> Column:
    ...


@overload
def transform(col: "ColumnOrName", f: Callable[[Column, Column], Column]) -> Column:
    ...


def transform(
    col: "ColumnOrName",
    f: Union[Callable[[Column], Column], Callable[[Column, Column], Column]],
) -> Column:
    """
    Returns an array of elements after applying a transformation to each element in the input array.

    .. versionadded:: 3.1.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        name of column or expression
    f : function
        a function that is applied to each element of the input array.
        Can take one of the following forms:

        - Unary ``(x: Column) -> Column: ...``
        - Binary ``(x: Column, i: Column) -> Column...``, where the second argument is
            a 0-based index of the element.

        and can use methods of :class:`~pyspark.sql.Column`, functions defined in
        :py:mod:`pyspark.sql.functions` and Scala ``UserDefinedFunctions``.
        Python ``UserDefinedFunctions`` are not supported
        (`SPARK-27052 <https://issues.apache.org/jira/browse/SPARK-27052>`__).

    Returns
    -------
    :class:`~pyspark.sql.Column`
        a new array of transformed elements.

    Examples
    --------
    >>> df = spark.createDataFrame([(1, [1, 2, 3, 4])], ("key", "values"))
    >>> df.select(transform("values", lambda x: x * 2).alias("doubled")).show()
    +------------+
    |     doubled|
    +------------+
    |[2, 4, 6, 8]|
    +------------+

    >>> def alternate(x, i):
    ...     return when(i % 2 == 0, x).otherwise(-x)
    ...
    >>> df.select(transform("values", alternate).alias("alternated")).show()
    +--------------+
    |    alternated|
    +--------------+
    |[1, -2, 3, -4]|
    +--------------+
    """
    raise NotImplementedError


def concat_ws(sep: str, *cols: "ColumnOrName") -> "Column":
    """
    Concatenates multiple input string columns together into a single string column,
    using the given separator.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    sep : str
        words separator.
    cols : :class:`~pyspark.sql.Column` or str
        list of columns to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        string of concatenated words.

    Examples
    --------
    >>> df = spark.createDataFrame([('abcd','123')], ['s', 'd'])
    >>> df.select(concat_ws('-', df.s, df.d).alias('s')).collect()
    [Row(s='abcd-123')]
    """
    cols = [_to_column(expr) for expr in cols]
    return _invoke_function(
        "concat_ws", ConstantExpression(sep), *cols
    )


def lower(col: "ColumnOrName") -> Column:
    """
    Converts a string expression to lower case.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        lower case values.

    Examples
    --------
    >>> df = spark.createDataFrame(["Spark", "PySpark", "Pandas API"], "STRING")
    >>> df.select(lower("value")).show()
    +------------+
    |lower(value)|
    +------------+
    |       spark|
    |     pyspark|
    |  pandas api|
    +------------+
    """
    return _invoke_function_over_columns("lower", col)


def ceil(col: "ColumnOrName") -> Column:
    """
    Computes the ceiling of the given value.

    .. versionadded:: 1.4.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(1)
    >>> df.select(ceil(lit(-0.1))).show()
    +----------+
    |CEIL(-0.1)|
    +----------+
    |         0|
    +----------+
    """
    return _invoke_function_over_columns("ceil", col)


def ceiling(col: "ColumnOrName") -> Column:
    return ceil(col)


def floor(col: "ColumnOrName") -> Column:
    """
    Computes the floor of the given value.

    .. versionadded:: 1.4.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        column to find floor for.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        nearest integer that is less than or equal to given value.

    Examples
    --------
    >>> df = spark.range(1)
    >>> df.select(floor(lit(2.5))).show()
    +----------+
    |FLOOR(2.5)|
    +----------+
    |         2|
    +----------+
    """
    return _invoke_function_over_columns("floor", col)


def abs(col: "ColumnOrName") -> Column:
    """
    Computes the absolute value.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(1)
    >>> df.select(abs(lit(-1))).show()
    +-------+
    |abs(-1)|
    +-------+
    |      1|
    +-------+
    """
    return _invoke_function_over_columns("abs", col)


def isnan(col: "ColumnOrName") -> Column:
    """An expression that returns true if the column is NaN.

    .. versionadded:: 1.6.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        True if value is NaN and False otherwise.

    Examples
    --------
    >>> df = spark.createDataFrame([(1.0, float('nan')), (float('nan'), 2.0)], ("a", "b"))
    >>> df.select("a", "b", isnan("a").alias("r1"), isnan(df.b).alias("r2")).show()
    +---+---+-----+-----+
    |  a|  b|   r1|   r2|
    +---+---+-----+-----+
    |1.0|NaN|false| true|
    |NaN|2.0| true|false|
    +---+---+-----+-----+
    """
    return _invoke_function_over_columns("isnan", col)


def isnull(col: "ColumnOrName") -> Column:
    """An expression that returns true if the column is null.

    .. versionadded:: 1.6.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        True if value is null and False otherwise.

    Examples
    --------
    >>> df = spark.createDataFrame([(1, None), (None, 2)], ("a", "b"))
    >>> df.select("a", "b", isnull("a").alias("r1"), isnull(df.b).alias("r2")).show()
    +----+----+-----+-----+
    |   a|   b|   r1|   r2|
    +----+----+-----+-----+
    |   1|NULL|false| true|
    |NULL|   2| true|false|
    +----+----+-----+-----+
    """
    return _invoke_function_over_columns("isnull", col)


def sqrt(col: "ColumnOrName") -> Column:
    """
    Computes the square root of the specified float value.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(1)
    >>> df.select(sqrt(lit(4))).show()
    +-------+
    |SQRT(4)|
    +-------+
    |    2.0|
    +-------+
    """
    return _invoke_function_over_columns("sqrt", col)


def greatest(*cols: "ColumnOrName") -> Column:
    """
    Returns the greatest value of the list of column names, skipping null values.
    This function takes at least 2 parameters. It will return null if all parameters are null.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        columns to check for gratest value.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        gratest value.

    Examples
    --------
    >>> df = spark.createDataFrame([(1, 4, 3)], ['a', 'b', 'c'])
    >>> df.select(greatest(df.a, df.b, df.c).alias("greatest")).collect()
    [Row(greatest=4)]
    """

    if len(cols) < 2:
        raise ValueError("greatest should take at least 2 columns")

    cols = [_to_column(expr) for expr in cols]
    return _invoke_function("greatest", *cols)


def least(*cols: "ColumnOrName") -> Column:
    """
    Returns the least value of the list of column names, skipping null values.
    This function takes at least 2 parameters. It will return null if all parameters are null.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    cols : :class:`~pyspark.sql.Column` or str
        column names or columns to be compared

    Returns
    -------
    :class:`~pyspark.sql.Column`
        least value.

    Examples
    --------
    >>> df = spark.createDataFrame([(1, 4, 3)], ['a', 'b', 'c'])
    >>> df.select(least(df.a, df.b, df.c).alias("least")).collect()
    [Row(least=1)]
    """
    if len(cols) < 2:
        raise ValueError("least should take at least 2 columns")

    cols = [_to_column(expr) for expr in cols]
    return _invoke_function("least", *cols)


def trim(col: "ColumnOrName") -> Column:
    """
    Trim the spaces from left end for the specified string value.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        left trimmed values.

    Examples
    --------
    >>> df = spark.createDataFrame(["   Spark", "Spark  ", " Spark"], "STRING")
    >>> df.select(ltrim("value").alias("r")).withColumn("length", length("r")).show()
    +-------+------+
    |      r|length|
    +-------+------+
    |  Spark|     5|
    |Spark  |     7|
    |  Spark|     5|
    +-------+------+
    """
    return _invoke_function_over_columns("trim", col)


def rtrim(col: "ColumnOrName") -> Column:
    """
    Trim the spaces from right end for the specified string value.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        right trimmed values.

    Examples
    --------
    >>> df = spark.createDataFrame(["   Spark", "Spark  ", " Spark"], "STRING")
    >>> df.select(rtrim("value").alias("r")).withColumn("length", length("r")).show()
    +--------+------+
    |       r|length|
    +--------+------+
    |   Spark|     8|
    |   Spark|     5|
    |   Spark|     6|
    +--------+------+
    """
    return _invoke_function_over_columns("rtrim", col)


def ltrim(col: "ColumnOrName") -> Column:
    """
    Trim the spaces from left end for the specified string value.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        left trimmed values.

    Examples
    --------
    >>> df = spark.createDataFrame(["   Spark", "Spark  ", " Spark"], "STRING")
    >>> df.select(ltrim("value").alias("r")).withColumn("length", length("r")).show()
    +-------+------+
    |      r|length|
    +-------+------+
    |  Spark|     5|
    |Spark  |     7|
    |  Spark|     5|
    +-------+------+
    """
    return _invoke_function_over_columns("ltrim", col)


def length(col: "ColumnOrName") -> Column:
    """Computes the character length of string data or number of bytes of binary data.
    The length of character data includes the trailing spaces. The length of binary data
    includes binary zeros.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        length of the value.

    Examples
    --------
    >>> spark.createDataFrame([('ABC ',)], ['a']).select(length('a').alias('length')).collect()
    [Row(length=4)]
    """
    return _invoke_function_over_columns("length", col)



import duckdb
import pytest

try:
    import pyarrow as pa

    can_run = True
except:
    can_run = False


def check_equal(duckdb_conn):
    true_result = duckdb_conn.execute("SELECT * from test").fetchall()
    duck_tbl = duckdb_conn.table("test")
    duck_from_arrow = duckdb_conn.from_arrow(duck_tbl.arrow())
    duck_from_arrow.create("testarrow")
    arrow_result = duckdb_conn.execute("SELECT * from testarrow").fetchall()
    assert arrow_result == true_result


class TestArrowFetch(object):
    def test_over_vector_size(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")
        for value in range(10000):
            duckdb_conn.execute("INSERT INTO  test VALUES (" + str(value) + ");")
        duckdb_conn.execute("INSERT INTO  test VALUES(NULL);")

        check_equal(duckdb_conn)

    def test_empty_table(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")

        check_equal(duckdb_conn)

    def test_over_vector_size(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()

        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")
        for value in range(10000):
            duckdb_conn.execute("INSERT INTO  test VALUES (" + str(value) + ");")
        duckdb_conn.execute("INSERT INTO  test VALUES(NULL);")

        check_equal(duckdb_conn)

    def test_table_nulls(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()

        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")
        duckdb_conn.execute("INSERT INTO  test VALUES(NULL);")

        check_equal(duckdb_conn)

    def test_table_without_nulls(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()

        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")
        duckdb_conn.execute("INSERT INTO  test VALUES(1);")

        check_equal(duckdb_conn)

    def test_table_with_prepared_statements(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()

        duckdb_conn.execute("CREATE TABLE test (a  INTEGER)")
        duckdb_conn.execute("PREPARE s1 AS INSERT INTO test VALUES ($1), ($2 / 2)")

        for value in range(10000):
            duckdb_conn.execute("EXECUTE s1(" + str(value) + "," + str(value * 2) + ");")

        check_equal(duckdb_conn)

    def test_to_arrow_chunk_size(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(3000);")
        relation = duckdb_cursor.table('t')
        arrow_tbl = relation.arrow()
        assert arrow_tbl['a'].num_chunks == 1
        arrow_tbl = relation.arrow(2048)
        assert arrow_tbl['a'].num_chunks == 2



import duckdb
import pytest
import datetime


class TestPythonResult(object):
    def test_result_closed(self, duckdb_cursor):
        connection = duckdb.connect('')
        cursor = connection.cursor()
        cursor.execute('CREATE TABLE integers (i integer)')
        cursor.execute('INSERT INTO integers VALUES (0),(1),(2),(3),(4),(5),(6),(7),(8),(9),(NULL)')
        rel = connection.table("integers")
        res = rel.aggregate("sum(i)").execute()
        res.close()
        with pytest.raises(duckdb.InvalidInputException, match='result closed'):
            res.fetchone()
        with pytest.raises(duckdb.InvalidInputException, match='result closed'):
            res.fetchall()
        with pytest.raises(duckdb.InvalidInputException, match='result closed'):
            res.fetchnumpy()
        with pytest.raises(duckdb.InvalidInputException, match='There is no query result'):
            res.fetch_arrow_table()
        with pytest.raises(duckdb.InvalidInputException, match='There is no query result'):
            res.fetch_arrow_reader(1)

    def test_result_describe_types(self, duckdb_cursor):
        connection = duckdb.connect('')
        cursor = connection.cursor()
        cursor.execute('CREATE TABLE test (i bool, j TIME, k VARCHAR)')
        cursor.execute("INSERT INTO test VALUES (TRUE, '01:01:01', 'bla' )")
        rel = connection.table("test")
        res = rel.execute()
        assert res.description == [
            ('i', 'bool', None, None, None, None, None),
            ('j', 'Time', None, None, None, None, None),
            ('k', 'STRING', None, None, None, None, None),
        ]

    def test_result_timestamps(self, duckdb_cursor):
        connection = duckdb.connect('')
        cursor = connection.cursor()
        cursor.execute(
            'CREATE TABLE IF NOT EXISTS timestamps (sec TIMESTAMP_S, milli TIMESTAMP_MS,micro TIMESTAMP_US, nano TIMESTAMP_NS );'
        )
        cursor.execute(
            "INSERT INTO timestamps VALUES ('2008-01-01 00:00:11','2008-01-01 00:00:01.794','2008-01-01 00:00:01.98926','2008-01-01 00:00:01.899268321' )"
        )

        rel = connection.table("timestamps")
        assert rel.execute().fetchall() == [
            (
                datetime.datetime(2008, 1, 1, 0, 0, 11),
                datetime.datetime(2008, 1, 1, 0, 0, 1, 794000),
                datetime.datetime(2008, 1, 1, 0, 0, 1, 989260),
                datetime.datetime(2008, 1, 1, 0, 0, 1, 899268),
            )
        ]

    def test_result_interval(self):
        connection = duckdb.connect()
        cursor = connection.cursor()
        cursor.execute('CREATE TABLE IF NOT EXISTS intervals (ivals INTERVAL)')
        cursor.execute("INSERT INTO intervals VALUES ('1 day'), ('2 second'), ('1 microsecond')")

        rel = connection.table("intervals")
        res = rel.execute()
        assert res.description == [('ivals', 'TIMEDELTA', None, None, None, None, None)]
        assert res.fetchall() == [
            (datetime.timedelta(days=1.0),),
            (datetime.timedelta(seconds=2.0),),
            (datetime.timedelta(microseconds=1.0),),
        ]

    def test_description_uuid(self):
        connection = duckdb.connect()
        connection.execute("select uuid();")
        connection.description



import numpy
import datetime
import pandas
import pytest
import duckdb
import re


def TestFile(name):
    import os

    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', name)
    return filename


class TestReadJSON(object):
    def test_read_json_columns(self):
        rel = duckdb.read_json(TestFile('example.json'), columns={'id': 'integer', 'name': 'varchar'})
        res = rel.fetchone()
        print(res)
        assert res == (1, 'O Brother, Where Art Thou?')

    def test_read_json_auto(self):
        rel = duckdb.read_json(TestFile('example.json'))
        res = rel.fetchone()
        print(res)
        assert res == (1, 'O Brother, Where Art Thou?')

    def test_read_json_maximum_depth(self):
        rel = duckdb.read_json(TestFile('example.json'), maximum_depth=4)
        res = rel.fetchone()
        print(res)
        assert res == (1, 'O Brother, Where Art Thou?')

    def test_read_json_sample_size(self):
        rel = duckdb.read_json(TestFile('example.json'), sample_size=2)
        res = rel.fetchone()
        print(res)
        assert res == (1, 'O Brother, Where Art Thou?')

    def test_read_json_format(self):
        # Wrong option
        with pytest.raises(duckdb.BinderException, match="format must be one of .* not 'test'"):
            rel = duckdb.read_json(TestFile('example.json'), format='test')

        rel = duckdb.read_json(TestFile('example.json'), format='unstructured')
        res = rel.fetchone()
        print(res)
        assert res == (
            [
                {'id': 1, 'name': 'O Brother, Where Art Thou?'},
                {'id': 2, 'name': 'Home for the Holidays'},
                {'id': 3, 'name': 'The Firm'},
                {'id': 4, 'name': 'Broadcast News'},
                {'id': 5, 'name': 'Raising Arizona'},
            ],
        )

    def test_read_json_records(self):
        # Wrong option
        with pytest.raises(duckdb.BinderException, match="""read_json requires "records" to be one of"""):
            rel = duckdb.read_json(TestFile('example.json'), records='none')

        rel = duckdb.read_json(TestFile('example.json'), records='true')
        res = rel.fetchone()
        print(res)
        assert res == (1, 'O Brother, Where Art Thou?')



import duckdb
import pandas as pd
import numpy
import pytest


def check_category_equal(category):
    df_in = pd.DataFrame(
        {
            'x': pd.Categorical(category, ordered=True),
        }
    )
    df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
    assert df_in.equals(df_out)


def check_result_list(category, res):
    for i in range(len(category)):
        assert category[i][0] == res[i]


def check_create_table(category):
    conn = duckdb.connect()

    conn.execute("PRAGMA enable_verification")
    df_in = pd.DataFrame({'x': pd.Categorical(category, ordered=True), 'y': pd.Categorical(category, ordered=True)})

    df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
    assert df_in.equals(df_out)

    conn.execute("CREATE TABLE t1 AS SELECT * FROM df_in")
    conn.execute("CREATE TABLE t2 AS SELECT * FROM df_in")

    # Check fetchall
    res = conn.execute("SELECT t1.x FROM t1").fetchall()
    check_result_list(res, category)

    # Do a insert to trigger string -> cat
    conn.execute("INSERT INTO t1 VALUES ('2','2')")

    res = conn.execute("SELECT x FROM t1 where x = '1'").fetchall()
    assert res == [('1',)]

    res = conn.execute("SELECT t1.x FROM t1 inner join t2 on (t1.x = t2.x)").fetchall()
    assert res == conn.execute("SELECT x FROM t1").fetchall()

    res = conn.execute("SELECT t1.x FROM t1 inner join t2 on (t1.x = t2.y)").fetchall()
    assert res == conn.execute("SELECT x FROM t1").fetchall()

    assert res == conn.execute("SELECT x FROM t1").fetchall()
    # Triggering the cast with ENUM as a src
    conn.execute("ALTER TABLE t1 ALTER x SET DATA TYPE VARCHAR")
    # We should be able to drop the table without any dependencies
    conn.execute("DROP TABLE t1")


class TestCategory(object):
    def test_category_simple(self, duckdb_cursor):
        df_in = pd.DataFrame({'float': [1.0, 2.0, 1.0], 'int': pd.Series([1, 2, 1], dtype="category")})

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        print(duckdb.query_df(df_in, "data", "SELECT * FROM data").fetchall())
        print(df_out['int'])
        assert numpy.all(df_out['float'] == numpy.array([1.0, 2.0, 1.0]))
        assert numpy.all(df_out['int'] == numpy.array([1, 2, 1]))

    def test_category_nulls(self, duckdb_cursor):
        df_in = pd.DataFrame({'int': pd.Series([1, 2, None], dtype="category")})
        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        print(duckdb.query_df(df_in, "data", "SELECT * FROM data").fetchall())
        assert df_out['int'][0] == 1
        assert df_out['int'][1] == 2
        assert numpy.isnan(df_out['int'][2])

    def test_category_string(self, duckdb_cursor):
        check_category_equal(['foo', 'bla', 'zoo', 'foo', 'foo', 'bla'])

    def test_category_string_null(self, duckdb_cursor):
        check_category_equal(['foo', 'bla', None, 'zoo', 'foo', 'foo', None, 'bla'])

    def test_category_string_null_bug_4747(self, duckdb_cursor):
        check_category_equal([str(i) for i in range(160)] + [None])

    def test_categorical_fetchall(self, duckdb_cursor):
        df_in = pd.DataFrame(
            {
                'x': pd.Categorical(['foo', 'bla', None, 'zoo', 'foo', 'foo', None, 'bla'], ordered=True),
            }
        )
        assert duckdb.query_df(df_in, "data", "SELECT * FROM data").fetchall() == [
            ('foo',),
            ('bla',),
            (None,),
            ('zoo',),
            ('foo',),
            ('foo',),
            (None,),
            ('bla',),
        ]

    def test_category_string_uint8(self, duckdb_cursor):
        category = []
        for i in range(10):
            category.append(str(i))
        check_create_table(category)

    def test_category_fetch_df_chunk(self, duckdb_cursor):
        con = duckdb.connect()
        categories = ['foo', 'bla', None, 'zoo', 'foo', 'foo', None, 'bla']
        result = categories * 256
        categories = result * 2
        df_result = pd.DataFrame(
            {
                'x': pd.Categorical(result, ordered=True),
            }
        )
        df_in = pd.DataFrame(
            {
                'x': pd.Categorical(categories, ordered=True),
            }
        )
        con.register("data", df_in)

        query = con.execute("SELECT * FROM data")
        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk.equals(df_result)

        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk.equals(df_result)

        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk.empty

    def test_category_mix(self, duckdb_cursor):
        df_in = pd.DataFrame(
            {
                'float': [1.0, 2.0, 1.0, 2.0, 1.0, 2.0, 1.0, 0.0],
                'x': pd.Categorical(['foo', 'bla', None, 'zoo', 'foo', 'foo', None, 'bla'], ordered=True),
            }
        )

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        assert df_out.equals(df_in)



import duckdb
import numpy
import pytest
from datetime import date, timedelta
import re
from conftest import NumpyPandas, ArrowPandas


# column count differs from bind
def evil1(df):
    if len(df) == 0:
        return df['col0'].to_frame()
    else:
        return df


class TestMap(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_evil_map(self, duckdb_cursor, pandas):
        testrel = duckdb.values([1, 2])
        with pytest.raises(duckdb.InvalidInputException, match='Expected 1 columns from UDF, got 2'):
            rel = testrel.map(evil1, schema={'i': str})
            df = rel.df()
            print(df)

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_map(self, duckdb_cursor, pandas):
        testrel = duckdb.values([1, 2])
        conn = duckdb_cursor
        conn.execute('CREATE TABLE t (a integer)')
        empty_rel = conn.table('t')

        newdf1 = testrel.map(lambda df: df['col0'].add(42).to_frame())
        newdf2 = testrel.map(lambda df: df['col0'].astype('string').to_frame())
        newdf3 = testrel.map(lambda df: df)

        # column type differs from bind
        def evil2(df):
            result = df.copy(deep=True)
            if len(result) == 0:
                result['col0'] = result['col0'].astype('double')
            return result

        # column name differs from bind
        def evil3(df):
            if len(df) == 0:
                df = df.rename(columns={"col0": "col42"})
            return df

        # does not return a df
        def evil4(df):
            return 42

        # straight up throws exception
        def evil5(df):
            raise TypeError

        def return_dataframe(df):
            return pandas.DataFrame({'A': [1]})

        def return_big_dataframe(df):
            return pandas.DataFrame({'A': [1] * 5000})

        def return_none(df):
            return None

        def return_empty_df(df):
            return pandas.DataFrame()

        with pytest.raises(duckdb.InvalidInputException, match='Expected 1 columns from UDF, got 2'):
            print(testrel.map(evil1).df())

        with pytest.raises(duckdb.InvalidInputException, match='UDF column type mismatch'):
            print(testrel.map(evil2).df())

        with pytest.raises(duckdb.InvalidInputException, match='UDF column name mismatch'):
            print(testrel.map(evil3).df())

        with pytest.raises(
            duckdb.InvalidInputException, match="Expected the UDF to return an object of type 'pandas.DataFrame'"
        ):
            print(testrel.map(evil4).df())

        with pytest.raises(duckdb.InvalidInputException):
            print(testrel.map(evil5).df())

        # not a function
        with pytest.raises(TypeError):
            print(testrel.map(42).df())

        # nothing passed to map
        with pytest.raises(TypeError):
            print(testrel.map().df())

        testrel.map(return_dataframe).df().equals(pandas.DataFrame({'A': [1]}))

        with pytest.raises(
            duckdb.InvalidInputException, match='UDF returned more than 2048 rows, which is not allowed.'
        ):
            testrel.map(return_big_dataframe).df()

        empty_rel.map(return_dataframe).df().equals(pandas.DataFrame({'A': []}))

        with pytest.raises(duckdb.InvalidInputException, match='No return value from Python function'):
            testrel.map(return_none).df()

        with pytest.raises(duckdb.InvalidInputException, match='Need a DataFrame with at least one column'):
            testrel.map(return_empty_df).df()

    def test_map_with_object_column(self, duckdb_cursor):
        def return_with_no_modification(df):
            return df

        # BLOB maps to 'object'
        # when a dataframe with 'object' column is returned, we use the content to infer the type
        # when the dataframe is empty, this results in NULL, which is not desirable
        # in this case we assume the returned type should be the same as the input type
        duckdb_cursor.values([b'1234']).map(return_with_no_modification).fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_isse_3237(self, duckdb_cursor, pandas):
        def process(rel):
            def mapper(x):
                dates = x['date'].to_numpy("datetime64[us]")
                days = x['days_to_add'].to_numpy("int")
                x["result1"] = pandas.Series(
                    [pandas.to_datetime(y[0]).date() + timedelta(days=y[1].item()) for y in zip(dates, days)],
                    dtype='datetime64[us]',
                )
                x["result2"] = pandas.Series(
                    [pandas.to_datetime(y[0]).date() + timedelta(days=-y[1].item()) for y in zip(dates, days)],
                    dtype='datetime64[us]',
                )
                return x

            rel = rel.map(mapper)
            rel = rel.project("*, datediff('day', date, result1) as one")
            rel = rel.project("*, datediff('day', date, result2) as two")
            rel = rel.project("*, IF(ABS(one) > ABS(two), one, two) as three")
            return rel

        df = pandas.DataFrame(
            {'date': pandas.Series([date(2000, 1, 1), date(2000, 1, 2)], dtype="datetime64[us]"), 'days_to_add': [1, 2]}
        )
        rel = duckdb.from_df(df)
        rel = process(rel)
        x = rel.fetchdf()
        assert x['days_to_add'].to_numpy()[0] == 1

    def test_explicit_schema(self):
        def cast_to_string(df):
            df['i'] = df['i'].astype(str)
            return df

        con = duckdb.connect()
        rel = con.sql('select i from range (10) tbl(i)')
        assert rel.types[0] == int
        mapped_rel = rel.map(cast_to_string, schema={'i': str})
        assert mapped_rel.types[0] == str

    def test_explicit_schema_returntype_mismatch(self):
        def does_nothing(df):
            return df

        con = duckdb.connect()
        rel = con.sql('select i from range(10) tbl(i)')
        # expects the mapper to return a string column
        rel = rel.map(does_nothing, schema={'i': str})
        with pytest.raises(
            duckdb.InvalidInputException, match=re.escape("UDF column type mismatch, expected [VARCHAR], got [BIGINT]")
        ):
            rel.fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_explicit_schema_name_mismatch(self, pandas):
        def renames_column(df):
            return pandas.DataFrame({'a': df['i']})

        con = duckdb.connect()
        rel = con.sql('select i from range(10) tbl(i)')
        rel = rel.map(renames_column, schema={'i': int})
        with pytest.raises(duckdb.InvalidInputException, match=re.escape('UDF column name mismatch')):
            rel.fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_explicit_schema_error(self, pandas):
        def no_op(df):
            return df

        con = duckdb.connect()
        rel = con.sql('select 42')
        with pytest.raises(
            duckdb.InvalidInputException,
            match=re.escape("Invalid Input Error: 'schema' should be given as a Dict[str, DuckDBType]"),
        ):
            rel.map(no_op, schema=[int])

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_returns_non_dataframe(self, pandas):
        def returns_series(df):
            return df.loc[:, 'i']

        con = duckdb.connect()
        rel = con.sql('select i, i as j from range(10) tbl(i)')
        with pytest.raises(
            duckdb.InvalidInputException,
            match=re.escape(
                "Expected the UDF to return an object of type 'pandas.DataFrame', found '<class 'pandas.core.series.Series'>' instead"
            ),
        ):
            rel = rel.map(returns_series)

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_explicit_schema_columncount_mismatch(self, pandas):
        def returns_subset(df):
            return pandas.DataFrame({'i': df.loc[:, 'i']})

        con = duckdb.connect()
        rel = con.sql('select i, i as j from range(10) tbl(i)')
        rel = rel.map(returns_subset, schema={'i': int, 'j': int})
        with pytest.raises(
            duckdb.InvalidInputException, match='Invalid Input Error: Expected 2 columns from UDF, got 1'
        ):
            rel.fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_pyarrow_df(self, pandas):
        # PyArrow backed dataframes only exist on pandas >= 2.0.0
        _ = pytest.importorskip("pandas", "2.0.0")

        def basic_function(df):
            # Create a pyarrow backed dataframe
            df = pandas.DataFrame({'a': [5, 3, 2, 1, 2]}).convert_dtypes(dtype_backend='pyarrow')
            return df

        con = duckdb.connect()
        with pytest.raises(duckdb.InvalidInputException):
            rel = con.sql('select 42').map(basic_function)



import duckdb
import pytest


class TestDBApiFetch(object):
    def test_multiple_fetch_one(self, duckdb_cursor):
        con = duckdb.connect()
        c = con.execute('SELECT 42')
        assert c.fetchone() == (42,)
        assert c.fetchone() is None
        assert c.fetchone() is None

    def test_multiple_fetch_all(self, duckdb_cursor):
        con = duckdb.connect()
        c = con.execute('SELECT 42')
        assert c.fetchall() == [(42,)]
        assert c.fetchall() == []
        assert c.fetchall() == []

    def test_multiple_fetch_many(self, duckdb_cursor):
        con = duckdb.connect()
        c = con.execute('SELECT 42')
        assert c.fetchmany(1000) == [(42,)]
        assert c.fetchmany(1000) == []
        assert c.fetchmany(1000) == []

    def test_multiple_fetch_df(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        con = duckdb.connect()
        c = con.execute('SELECT 42::BIGINT AS a')
        pd.testing.assert_frame_equal(c.df(), pd.DataFrame.from_dict({'a': [42]}))
        assert c.df() is None
        assert c.df() is None

    def test_multiple_fetch_arrow(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        arrow = pytest.importorskip("pyarrow")
        con = duckdb.connect()
        c = con.execute('SELECT 42::BIGINT AS a')
        table = c.arrow()
        df = table.to_pandas()
        pd.testing.assert_frame_equal(df, pd.DataFrame.from_dict({'a': [42]}))
        assert c.arrow() is None
        assert c.arrow() is None

    def test_multiple_close(self, duckdb_cursor):
        con = duckdb.connect()
        c = con.execute('SELECT 42')
        c.close()
        c.close()
        c.close()
        with pytest.raises(duckdb.InvalidInputException, match='No open result set'):
            c.fetchall()

    def test_multiple_fetch_all_relation(self, duckdb_cursor):
        res = duckdb_cursor.query('SELECT 42')
        assert res.fetchall() == [(42,)]
        assert res.fetchall() == [(42,)]
        assert res.fetchall() == [(42,)]

    def test_multiple_fetch_many_relation(self, duckdb_cursor):
        res = duckdb_cursor.query('SELECT 42')
        assert res.fetchmany(10000) == [(42,)]
        assert res.fetchmany(10000) == []
        assert res.fetchmany(10000) == []

    def test_fetch_one_relation(self, duckdb_cursor):
        res = duckdb_cursor.query('SELECT * FROM range(3)')
        assert res.fetchone() == (0,)
        assert res.fetchone() == (1,)
        assert res.fetchone() == (2,)
        assert res.fetchone() is None
        assert res.fetchone() is None
        assert res.fetchone() is None
        assert res.fetchall() == []
        # we can execute explicitly to reset the result
        res.execute()
        assert res.fetchone() == (0,)
        res.execute()
        assert res.fetchone() == (0,)
        assert res.fetchone() == (1,)
        assert res.fetchone() == (2,)
        assert res.fetchone() is None



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.types import Row
from duckdb.experimental.spark.sql import functions as F


class TestStringFunctions(object):
    def test_length(self, spark):
        data = [
            ("firstRowFirstColumn",),
            ("2ndRowFirstColumn",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("length", F.length(F.col("firstColumn")))
        res = df.select("length").collect()
        assert res == [
            Row(length=19),
            Row(length=17),
        ]

    def test_upper(self, spark):
        data = [
            ("firstRowFirstColumn",),
            ("2ndRowFirstColumn",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("upper", F.upper(F.col("firstColumn")))
        res = df.select("upper").collect()
        assert res == [
            Row(upper="FIRSTROWFIRSTCOLUMN"),
            Row(upper="2NDROWFIRSTCOLUMN"),
        ]

    def test_lower(self, spark):
        data = [
            ("firstRowFirstColumn",),
            ("2ndRowFirstColumn",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("lower", F.lower(F.col("firstColumn")))
        res = df.select("lower").collect()
        assert res == [
            Row(lower="firstrowfirstcolumn"),
            Row(lower="2ndrowfirstcolumn"),
        ]

    def test_trim(self, spark):
        data = [
            (" firstRowFirstColumn ",),
            (" 2ndRowFirstColumn ",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("trimmed", F.trim(F.col("firstColumn")))
        res = df.select("trimmed").collect()
        assert res == [
            Row(trimmed="firstRowFirstColumn"),
            Row(trimmed="2ndRowFirstColumn"),
        ]

    def test_ltrim(self, spark):
        data = [
            (" firstRowFirstColumn ",),
            (" 2ndRowFirstColumn ",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("ltrimmed", F.ltrim(F.col("firstColumn")))
        res = df.select("ltrimmed").collect()
        assert res == [
            Row(ltrimmed="firstRowFirstColumn "),
            Row(ltrimmed="2ndRowFirstColumn "),
        ]

    def test_rtrim(self, spark):
        data = [
            (" firstRowFirstColumn ",),
            (" 2ndRowFirstColumn ",),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("rtrimmed", F.rtrim(F.col("firstColumn")))
        res = df.select("rtrimmed").collect()
        assert res == [
            Row(rtrimmed=" firstRowFirstColumn"),
            Row(rtrimmed=" 2ndRowFirstColumn"),
        ]



from typing import Dict, Optional, cast

from ..utils import ErrorClassesReader

class PySparkException(Exception):
    """
    Base Exception for handling errors generated from PySpark.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        # The error class, decides the message format, must be one of the valid options listed in 'error_classes.py'
        error_class: Optional[str] = None,
        # The dictionary listing the arguments specified in the message (or the error_class)
        message_parameters: Optional[Dict[str, str]] = None,
    ):
        # `message` vs `error_class` & `message_parameters` are mutually exclusive.
        assert (message is not None and (error_class is None and message_parameters is None)) or (
            message is None and (error_class is not None and message_parameters is not None)
        )

        self.error_reader = ErrorClassesReader()

        if message is None:
            self.message = self.error_reader.get_error_message(
                cast(str, error_class), cast(Dict[str, str], message_parameters)
            )
        else:
            self.message = message

        self.error_class = error_class
        self.message_parameters = message_parameters

    def getErrorClass(self) -> Optional[str]:
        """
        Returns an error class as a string.

        .. versionadded:: 3.4.0

        See Also
        --------
        :meth:`PySparkException.getMessageParameters`
        :meth:`PySparkException.getSqlState`
        """
        return self.error_class

    def getMessageParameters(self) -> Optional[Dict[str, str]]:
        """
        Returns a message parameters as a dictionary.

        .. versionadded:: 3.4.0

        See Also
        --------
        :meth:`PySparkException.getErrorClass`
        :meth:`PySparkException.getSqlState`
        """
        return self.message_parameters

    def getSqlState(self) -> None:
        """
        Returns an SQLSTATE as a string.

        Errors generated in Python have no SQLSTATE, so it always returns None.

        .. versionadded:: 3.4.0

        See Also
        --------
        :meth:`PySparkException.getErrorClass`
        :meth:`PySparkException.getMessageParameters`
        """
        return None

    def __str__(self) -> str:
        if self.getErrorClass() is not None:
            return f"[{self.getErrorClass()}] {self.message}"
        else:
            return self.message

class AnalysisException(PySparkException):
    """
    Failed to analyze a SQL query plan.
    """


class SessionNotSameException(PySparkException):
    """
    Performed the same operation on different SparkSession.
    """


class TempTableAlreadyExistsException(AnalysisException):
    """
    Failed to create temp view since it is already exists.
    """


class ParseException(AnalysisException):
    """
    Failed to parse a SQL command.
    """


class IllegalArgumentException(PySparkException):
    """
    Passed an illegal or inappropriate argument.
    """


class ArithmeticException(PySparkException):
    """
    Arithmetic exception thrown from Spark with an error class.
    """


class UnsupportedOperationException(PySparkException):
    """
    Unsupported operation exception thrown from Spark with an error class.
    """


class ArrayIndexOutOfBoundsException(PySparkException):
    """
    Array index out of bounds exception thrown from Spark with an error class.
    """


class DateTimeException(PySparkException):
    """
    Datetime exception thrown from Spark with an error class.
    """


class NumberFormatException(IllegalArgumentException):
    """
    Number format exception thrown from Spark with an error class.
    """


class StreamingQueryException(PySparkException):
    """
    Exception that stopped a :class:`StreamingQuery`.
    """


class QueryExecutionException(PySparkException):
    """
    Failed to execute a query.
    """


class PythonException(PySparkException):
    """
    Exceptions thrown from Python workers.
    """


class SparkRuntimeException(PySparkException):
    """
    Runtime exception thrown from Spark with an error class.
    """


class SparkUpgradeException(PySparkException):
    """
    Exception thrown because of Spark upgrade.
    """


class UnknownException(PySparkException):
    """
    None of the above exceptions.
    """


class PySparkValueError(PySparkException, ValueError):
    """
    Wrapper class for ValueError to support error classes.
    """


class PySparkTypeError(PySparkException, TypeError):
    """
    Wrapper class for TypeError to support error classes.
    """


class PySparkAttributeError(PySparkException, AttributeError):
    """
    Wrapper class for AttributeError to support error classes.
    """


class PySparkRuntimeError(PySparkException, RuntimeError):
    """
    Wrapper class for RuntimeError to support error classes.
    """


class PySparkAssertionError(PySparkException, AssertionError):
    """
    Wrapper class for AssertionError to support error classes.
    """


class PySparkNotImplementedError(PySparkException, NotImplementedError):
    """
    Wrapper class for NotImplementedError to support error classes.
    """



import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas

pyarrow = pytest.importorskip('pyarrow')


@pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
def test_6796(pandas):
    conn = duckdb.connect()
    input_df = pandas.DataFrame({"foo": ["bar"]})
    conn.register("input_df", input_df)

    query = """
	select * from input_df
	union all
	select * from input_df
	"""

    # fetching directly into Pandas works
    res_df = conn.execute(query).fetch_df()
    res_arrow = conn.execute(query).fetch_arrow_table()

    df_arrow_table = pyarrow.Table.from_pandas(res_df)

    result_1 = conn.execute("select * from df_arrow_table order by all").fetchall()

    result_2 = conn.execute("select * from res_arrow order by all").fetchall()

    assert result_1 == result_2



import sys

import duckdb
import pytest

pa = pytest.importorskip("pyarrow")
adbc_driver_manager = pytest.importorskip("adbc_driver_manager")

try:
    adbc_driver_duckdb = pytest.importorskip("adbc_driver_duckdb.dbapi")
    con = adbc_driver_duckdb.connect()
except adbc_driver_manager.InternalError as e:
    pytest.skip(
        f"'duckdb_adbc_init' was not exported in this install, try running 'python3 setup.py install': {e}",
        allow_module_level=True,
    )


class TestADBCConnectionGetInfo(object):
    def test_connection_basic(self):
        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            cursor.execute("select 42")
            res = cursor.fetchall()
            assert res == [(42,)]

    def test_connection_get_info_all(self):
        con = adbc_driver_duckdb.connect()
        adbc_con = con.adbc_connection
        res = adbc_con.get_info()
        reader = pa.RecordBatchReader._import_from_c(res.address)
        table = reader.read_all()
        values = table["info_value"]

        expected_result = pa.array(
            [
                "duckdb",
                "v" + duckdb.__version__,  # don't hardcode this, as it will change every version
                "ADBC DuckDB Driver",
                "(unknown)",
                "(unknown)",
            ],
            type=pa.string(),
        )

        assert values.num_chunks == 1
        chunk = values.chunk(0)
        string_values = chunk.field(0)
        assert string_values == expected_result

    def test_empty_result(self):
        con = adbc_driver_duckdb.connect()
        adbc_con = con.adbc_connection
        res = adbc_con.get_info([1337])
        reader = pa.RecordBatchReader._import_from_c(res.address)
        table = reader.read_all()
        values = table["info_value"]

        # Because all the codes we asked for were unrecognized, the result set is empty
        assert values.num_chunks == 0

    def test_unrecognized_codes(self):
        con = adbc_driver_duckdb.connect()
        adbc_con = con.adbc_connection
        res = adbc_con.get_info([0, 1000, 4, 2000])
        reader = pa.RecordBatchReader._import_from_c(res.address)
        table = reader.read_all()
        values = table["info_value"]

        expected_result = pa.array(["duckdb", "(unknown)"], type=pa.string())

        assert values.num_chunks == 1
        chunk = values.chunk(0)
        string_values = chunk.field(0)
        assert string_values == expected_result
